from datetime import date, datetime, time
from typing import get_args

from dotenv import load_dotenv
from tariff_fetch.arcadia.api import ArcadiaSignalAPI
from tariff_fetch.arcadia.schema.common import RateChargeClass
from tariff_fetch.urdb.arcadia.energyschedule import get_raw_bands_at_datetime
from tariff_fetch.urdb.arcadia.library import Library
from tariff_fetch.urdb.arcadia.scenario import Scenario


def main():
    _ = load_dotenv()
    master_tariff_id = 80824
    scenario = Scenario(master_tariff_id, 2025, apply_percentages=True, charge_classes=set(get_args(RateChargeClass)))
    api = ArcadiaSignalAPI()
    library = Library(api)
    result = get_raw_bands_at_datetime(scenario, library, datetime.combine(date(2025, 1, 15), time(minute=30)))
    print(result)

    pass


if __name__ == "__main__":
    main()



from enum import Enum
from typing import NamedTuple


class Provider(str, Enum):
    GENABILITY = "Genability"
    OPENEI = "OpenEI"
    RATEACUITY = "RateAcuity"


class StateCode(str, Enum):
    AL = "al"
    AK = "ak"
    AZ = "az"
    AR = "ar"
    CA = "ca"
    CO = "co"
    CT = "ct"
    DE = "de"
    FL = "fl"
    GA = "ga"
    HI = "hi"
    ID = "id"
    IL = "il"
    IN = "in"
    IA = "ia"
    KS = "ks"
    KY = "ky"
    LA = "la"
    ME = "me"
    MD = "md"
    MA = "ma"
    MI = "mi"
    MN = "mn"
    MS = "ms"
    MO = "mo"
    MT = "mt"
    NE = "ne"
    NV = "nv"
    NH = "nh"
    NJ = "nj"
    NM = "nm"
    NY = "ny"
    NC = "nc"
    ND = "nd"
    OH = "oh"
    OK = "ok"
    OR = "or"
    PA = "pa"
    RI = "ri"
    SC = "sc"
    SD = "sd"
    TN = "tn"
    TX = "tx"
    UT = "ut"
    VT = "vt"
    VA = "va"
    WA = "wa"
    WV = "wv"
    WI = "wi"
    WY = "wy"
    DC = "dc"


class Utility(NamedTuple):
    eia_id: int
    name: str



from datetime import datetime
from typing import Annotated, NotRequired

from pydantic import BeforeValidator
from pydantic.alias_generators import to_camel
from typing_extensions import TypedDict

from .common import RateChargeClass, RateTransactionType, RateUnit, TariffChargePeriod, TariffChargeType
from .season import SeasonExtended, SeasonStandard
from .territory import TerritoryStandard
from .timeofuse import TimeOfUseExtended, TimeOfUseStandard
from .validators import comma_separated_str


class TariffRateBand(TypedDict):
    tariff_rate_band_id: int
    """Unique Arcadia ID (primary key) for each Band"""
    tariff_rate_id: int
    """ID of the rate this band belongs to (foreign key)"""
    rate_sequence_number: int
    """This bands position in the bands for its rate"""
    has_consumption_limit: bool
    """True indicates that this has banded consumption"""
    consumption_upper_limit: NotRequired[float]
    """indicates the upper consumption limit of this band"""
    has_demand_limit: bool
    """True indicates that this has banded demand"""
    demand_upper_limit: NotRequired[float]
    """indicates the upper demand limit of this band"""
    has_property_limit: bool
    """true indicates that this has a limit based on a property"""
    property_upper_limit: NotRequired[float]
    """indicates the upper limit of this band"""
    prev_upper_limit: float | None
    """The upper limit of the previous rate band for this rate"""
    applicability_value: NotRequired[str]
    """indicates the value of applicability property that qualifies for this rate"""
    calculation_factor: NotRequired[float]
    """A factor to be applied to the cost of the rate."""
    rate_amount: float
    """Charge amount for this band"""
    rate_unit: RateUnit
    """
    Possible values are:

    COST_PER_UNIT - rate amount multiplied by the number of units
    PERCENTAGE - percentage of a value (e.g. percentage of overall bill)
    """
    is_credit: bool
    """When true this band is a credit amount (reduces the bill)"""
    applicability_formula: NotRequired[str]


class TariffRateMinimalFields(TypedDict):
    master_tariff_rate_id: int | None
    has_applicability_formula: bool
    tariff_rate_id: int
    """Unique Arcadia ID (primary key) for each tariff rate"""
    tariff_id: int
    """Associates the rate with a tariff (foreign key)"""
    rider_id: NotRequired[int]
    """Tariff ID of the rider attached to this tariff version."""
    tariff_sequence_number: int
    """Sequence of this rate in the tariff, for display purposes only (e.g. this is the order on the bill)"""
    rate_group_name: str
    """Name of the group this rate belongs to"""
    rate_name: str | None
    """Name of this rate. If None, `rate_group_name` is used."""


class TariffRateStandardFields(TypedDict):
    from_date_time: datetime | None
    """Indicates the rate's effective date is not the same as that of its tariff"""
    to_date_time: datetime | None
    """Indicates the rates end date is not the same as that of its tariff"""
    charge_type: TariffChargeType | None
    """
    Possible values are:
        FIXED_PRICE - a fixed charge for the period
        CONSUMPTION_BASED - based on quantity used (e.g. kW/h)
        DEMAND_BASED - based on the peak demand (e.g. kW)
        QUANTITY - a rate per number of items (e.g. $5 per street light)
        FORMULA - a rate that has a specific or custom formula
        MINIMUM - a minimum amount that the LSE will charge you, overriding lower pre-tax charges
        MAXIMUM - a maximum amount that the LSE will charge you, overriding higher pre-tax charges
        TAX - a percentage tax rate which is applied to the sum of all of the other charges on a bill
    """
    charge_class: NotRequired[Annotated[list[RateChargeClass], BeforeValidator(comma_separated_str)]]
    """indicates what class(es) of charges this rate is for.

    Values include:
        SUPPLY - Energy-related charges.
        TRANSMISSION - Transmission level delivery charges.
        DISTRIBUTION - Distribution level (last mile) delivery charges of moving electricity into your home or business.
        TAX - Tax surcharges that appear in the utility tariff document.
        CONTRACTED - Charges that get replaced or overridden when you pass the retail (contracted) energy supply rate in the calculation.
        USER_ADJUSTED - Additional or custom rates you can add to a public or private tariff.
            An example would be a local tax rate which Arcadia does not model but you would want included.
        AFTER_TAX - Charges which apply post utility and other taxes.
            A good example would be the California Climate credit which is applied to the bill after the taxes are applied to the bill subtotal.
        OTHER - Charges which cannot be classified in any of the above buckets. This is very rare.
        NON_BYPASSABLE - Charges which cannot be offset by credits (usually Net Metering)
    """
    charge_period: TariffChargePeriod
    """
    Indicates what period this charge is calculated for.
    This is usually the same as the billing period (and is usually monthly) but can be other intervals.
    """
    quantity_key: NotRequired[str | None]
    """ When not null, the property that defines the type of quantity this rate applies to.
        (e.g. billingMeter : property which defines the number of billing meters the rate will apply to)
    """
    applicability_key: NotRequired[str]
    """ defines the eligibility criteria for this rate. (e.g. connectionType : property which defines how the service is connected to the grid) """
    variable_limit_key: NotRequired[str]
    """defines the variable which determines the upper limit(s) of this rate

    e.g. demandMultiplierTierswithkWhTiers2416: property which uses the demand value to drive the consumption limits
    """
    variable_rate_key: NotRequired[str]
    variable_rate_sub_key: NotRequired[str]
    """this is the name of the property that defines the variable rate

    e.g massachusettsResidentialRetailPrevailingRates : property which provides the regional prevailing residential supply rate for Massachusetts
    """
    variable_factor_key: NotRequired[str]
    """The name of the property that defines the variable factor to apply to this rate.

    e.g billingPeriodProrationFactor: property which defines a prorated number of billing days
    """
    rate_bands: list[TariffRateBand]

    # These are fields with unknown purpose
    edge_predominance: NotRequired[str]
    proration_rules: NotRequired[list[str]]


class TariffRateExtendedFields(TypedDict):
    rider_tariff_id: NotRequired[int]
    """Tariff ID of the rider attached to this tariff version. """
    tariff_book_sequence_number: int | None
    """Sequence of this rate in the tariff source document, if it differs from tariff_sequence_number"""
    tariff_book_rate_group_name: str | None
    """Name of the group this rate belongs to in the tariff source document, if it differs from rate_group_name"""
    tariff_book_rate_name: str | None
    """Name of this rate in the tariff source document, if it differs from rate_name"""
    transaction_type: RateTransactionType
    """
    Indicates whether this rate is BUY (charge when importing from the grid, no credit when exporting),
    SELL (credit when exporting to the grid, no charge when importing),
    or NET (charge when importing, credit when exporting) with imports and exports resolved according the the chargePeriod.
    BUY_IMPORT (charge only when importing) and
    SELL_EXPORT (credit only when exporting) indicate that imports and exports are resolved in real-time
    (instantaneous netting).
    """


class TariffRateMinimal(TariffRateMinimalFields):
    __pydantic_config__ = {"alias_generator": to_camel, "extra": "forbid"}  # pyright: ignore[reportGeneralTypeIssues, reportUnannotatedClassAttribute]  # noqa


class TariffRateStandard(TariffRateMinimalFields, TariffRateStandardFields):
    __pydantic_config__ = {"alias_generator": to_camel, "extra": "forbid"}  # pyright: ignore[reportGeneralTypeIssues, reportUnannotatedClassAttribute]  # noqa
    time_of_use: NotRequired[TimeOfUseStandard]
    """The time period this rate applies to. Only used for TOU rates"""
    territory: NotRequired[TerritoryStandard]
    """Only populated when this rate applies to a different region than the whole tariff (e.g. California Baseline Regions)."""
    season: NotRequired[SeasonStandard]
    """The season this rate applies to. Only used for seasonal rates"""


class TariffRateExtended(TariffRateMinimalFields, TariffRateStandardFields, TariffRateExtendedFields):
    __pydantic_config__ = {"alias_generator": to_camel, "extra": "forbid"}  # pyright: ignore[reportGeneralTypeIssues, reportUnannotatedClassAttribute]  # noqa
    time_of_use: NotRequired[TimeOfUseExtended]
    """The time period this rate applies to. Only used for TOU rates"""
    territory: NotRequired[TerritoryStandard]
    """Only populated when this rate applies to a different region than the whole tariff (e.g. California Baseline Regions)."""
    season: NotRequired[SeasonExtended]
    """The season this rate applies to. Only used for seasonal rates"""



import re

from .types import BandDeterminant, ConsumptionRateDeterminant

_DATE_COL_RE = re.compile(r"^(0?[1-9]|1[0-2])/(0?[1-9]|[12]\d|3[01])/(\d{4})$")


def is_date_column_name(column_name: str) -> bool:
    return bool(_DATE_COL_RE.match(column_name))


def kwh_multiplier(determinant: ConsumptionRateDeterminant | BandDeterminant) -> float:
    match determinant:
        case "per therm" | "therms":
            return 29.3
        case "per ccf" | "ccf":
            return 29.31
        case "cubic feet":
            return 0.2931



"""Build basic URDB metadata fields from the selected Arcadia tariff version."""

from datetime import date

from tariff_fetch.urdb.schema import URDBRate

from .library import Library
from .scenario import Scenario


def build_metadata(scenario: Scenario, library: Library) -> URDBRate:
    """Map the scenario tariff's utility and tariff names into URDB metadata fields."""

    master_tariff_id = scenario.master_tariff_id
    tariff = library.tariffs.get_tariff_at_date(master_tariff_id, date(scenario.year, 1, 1))
    return {"label": tariff["lse_code"], "utility": tariff["lse_name"], "name": tariff["tariff_name"], "country": "USA"}



from __future__ import annotations

from typing import cast

from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement
from typing_extensions import TypedDict

from .schema import Section


class TableJson(TypedDict):
    title: str
    columns: list[str]
    values: list[dict[str, str]]


class SectionJson(TypedDict):
    section: str | None
    tables: list[TableJson]


def _headers_from_table(table: WebElement) -> list[str]:
    ths = table.find_elements(By.CSS_SELECTOR, "thead th")  # pyright: ignore[reportUnknownMemberType]
    result: list[str] = []
    for th in ths:
        links = th.find_elements(By.TAG_NAME, "a")  # pyright: ignore[reportUnknownMemberType]
        result.append(links[0].text if links else th.text)
    return result


def _rows_from_table(table: WebElement) -> list[list[str]]:
    rows: list[list[str]] = []
    for tr in table.find_elements(By.CSS_SELECTOR, "tbody tr"):  # pyright: ignore[reportUnknownMemberType]
        tds = tr.find_elements(By.TAG_NAME, "td")  # pyright: ignore[reportUnknownMemberType]
        rows.append([td.text for td in tds])
    return rows


def _table_json(table: WebElement) -> TableJson | None:
    columns = _headers_from_table(table)
    if not columns:
        return None
    rows = _rows_from_table(table)
    title = columns[0]

    values = []

    for row in rows:
        v = {}
        for c, r in zip(columns, row, strict=False):
            v[c] = r
        values.append(v)  # pyright: ignore[reportUnknownMemberType]

    return TableJson(
        {
            "title": title,
            "columns": columns,
            "values": values,
        }
    )


def sections_to_json(driver: WebDriver) -> list[Section]:
    seq = driver.find_elements(By.CSS_SELECTOR, "h3, table.eamwebgrid-table")

    sections: list[SectionJson] = []
    current: SectionJson = {"section": None, "tables": []}
    for el in seq:
        tag = el.tag_name.lower()
        if tag == "h3":
            if current["section"] is not None or current["tables"]:
                sections.append(current)
            current = {"section": el.text, "tables": []}
        else:
            table = _table_json(el)
            if table:
                current["tables"].append(table)

    if current["section"] is not None or current["tables"]:
        sections.append(current)

    # Schema is not yet complete, ignore validation
    return cast(list[Section], sections)
    # ta = TypeAdapter(list[Section])

    # return ta.validate_python(sections, by_alias=True)



from datetime import date
from typing import Literal, Required, get_args

from typing_extensions import TypedDict, TypeIs

BandDeterminant = Literal["ccf", "therms", "cubic feet"]
ConsumptionRateDeterminant = Literal["per ccf", "per therm"]
PercentRateDeterminant = Literal["percent"]
FixedRateDeterminant = Literal["per month", "per month per location", "per month per meter", "per meter per month"]
RateDeterminant = Literal[ConsumptionRateDeterminant, PercentRateDeterminant, FixedRateDeterminant]
KnownRateDeterminant = RateDeterminant


class Row(TypedDict, total=False):
    rate: Required[str]
    effective_date: date | None
    season: str | None
    min_therms: float | None
    max_therms: float | None
    start: Required[float | None]
    end: Required[float | None]
    location: str
    determinant: BandDeterminant | None
    rate_determinant: Required[RateDeterminant]


def is_band_determinant(value: str) -> TypeIs[BandDeterminant]:
    return value in get_args(BandDeterminant)


def is_consumption_rate_determinant(value: str) -> TypeIs[ConsumptionRateDeterminant]:
    return value in get_args(ConsumptionRateDeterminant)


def is_rate_determinant(value: str) -> TypeIs[RateDeterminant]:
    return value in get_args(RateDeterminant)



import logging
import os
from collections.abc import Generator
from contextlib import contextmanager
from tempfile import TemporaryDirectory
from typing import NamedTuple

from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

logger = logging.getLogger(__name__)


class AuthorizationError(Exception):
    pass


class ScrapingContext(NamedTuple):
    driver: webdriver.Chrome
    download_path: str


@contextmanager
def create_context() -> Generator[ScrapingContext]:
    with TemporaryDirectory() as temp_dir:
        driver = create_driver_(temp_dir)
        driver.set_window_size(1920, 1080)  # pyright: ignore[reportUnknownMemberType]
        try:
            yield ScrapingContext(driver, temp_dir)
        except Exception as e:
            _ = driver.save_screenshot("selenium_error.png")  # pyright: ignore[reportUnknownMemberType]
            raise e from None


def create_driver_(download_path: str) -> webdriver.Chrome:
    options = Options()
    options.add_argument("--headless")
    options.add_argument("--no-sandbox")
    options.add_argument("--disable-dev-shm-usage")
    options.add_argument(
        "user-agent=Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/113.0.0.0 Safari/537.36"
    )
    options.add_experimental_option(  # pyright: ignore[reportUnknownMemberType]
        "prefs",
        {
            "download.default_directory": download_path,
            "download.prompt_for_download": False,
            "directory_upgrade": True,
        },
    )
    service = Service(log_path=os.devnull)
    return webdriver.Chrome(options=options, service=service)


def login(driver: webdriver.Chrome, email_address: str, password: str):
    logger.info("Logging in")
    driver.get("https://secure.rateacuity.com/RateAcuityPortal/Account/Login")
    WebDriverWait(driver, 10).until(EC.presence_of_element_located((By.ID, "UserName"))).send_keys(email_address)
    WebDriverWait(driver, 10).until(EC.presence_of_element_located((By.ID, "Password"))).send_keys(password)
    WebDriverWait(driver, 10).until(
        EC.element_to_be_clickable((By.XPATH, "//input[@type='submit' and @value='Log in']"))
    ).click()
    driver.execute_script("Object.defineProperty(navigator, 'webdriver', {get: () => undefined})")  # pyright: ignore[reportUnknownMemberType]



import json
import os
from datetime import date
from pathlib import Path
from typing import cast

import questionary
from dotenv import load_dotenv

# from tariff_fetch.genability.lse import get_lses_page
# from tariff_fetch.genability.tariffs import CustomerClass, TariffType, tariffs_paginate
from tariff_fetch.arcadia.api import ArcadiaSignalAPI
from tariff_fetch.arcadia.schema.common import CustomerClass, TariffType
from tariff_fetch.arcadia.schema.tariff import TariffExtended
from tariff_fetch.urdb.arcadia.build import build_urdb
from tariff_fetch.urdb.arcadia.prompts import prompt_charge_classes
from tariff_fetch.urdb.arcadia.scenario import Scenario
from tariff_fetch.urdb.schema import URDBRate

from . import console, prompt_filename
from .types import Utility


def process_genability(utility: Utility, output_folder: Path, year: int):
    _ = load_dotenv()
    if not os.getenv("ARCADIA_APP_ID"):
        console.print("[b]ARCADIA_APP_ID[/] environment variable is not set.")
    if not os.getenv("ARCADIA_APP_KEY"):
        console.print("[b]ARCADIA_APP_KEY[/] environment variable is not set.")
    if not (os.getenv("ARCADIA_APP_ID") and os.getenv("ARCADIA_APP_KEY")):
        console.print("Cannot use Arcadia API due to missing credentials")
        _ = console.input("Press enter to proceed...")
        return
    api = ArcadiaSignalAPI()

    lse_id = _find_utility_lse_id(api, utility)
    if lse_id is None:
        return

    if not (customer_classes := _select_customer_classes()):
        return

    if not (tariff_types := _select_tariff_types()):
        return

    if not (tariffs := _select_tariffs(api, lse_id, customer_classes, tariff_types, year)):
        console.print("[red]No tariffs found[/]")
        _ = console.input("Press enter to proceed...")
        return

    api_results = _fetch_tariffs(api, tariffs, year)
    results: list[URDBRate] = []

    # tariff_ids = {id_ for _, id_ in tariffs}

    apply_percentages = cast(bool | None, questionary.confirm("Apply percentage rates?").ask())
    if apply_percentages is None:
        return

    for tariff in api_results:
        tariff_name = tariff["tariff_name"]
        console.print(f"Creating scenario for tariff: [blue]{tariff_name}[/]")
        charge_classes = prompt_charge_classes()
        if charge_classes is None:
            continue
        scenario = Scenario(
            tariff["master_tariff_id"],
            year=year,
            apply_percentages=apply_percentages,
            charge_classes=charge_classes,
        )
        urdb_tariff = build_urdb(api, scenario)
        urdb_tariff["name"] = _prompt_tariff_name(urdb_tariff.get("name", ""))

        results.append(urdb_tariff)
        # results.append(build_urdb(tariff, api, scenario))

    suggested_filename = f"arcadia_{utility.name}.urdb.{year}."

    if not (filename := prompt_filename(output_folder, suggested_filename, "json")):
        return

    filename.parent.mkdir(exist_ok=True)
    _ = filename.write_text(json.dumps({"items": results}, indent=2))
    console.print(f"Wrote [blue]{len(results)}[/] records to {filename}")


def _find_utility_lse_id(api: ArcadiaSignalAPI, utility: Utility) -> int | None:
    with console.status("Fetching lses..."):
        lses = api.lses.get_page(
            fields="min",
            search_on=["code"],
            search=str(utility.eia_id),
            starts_with=True,
            ends_with=True,
        )["results"]
    if len(lses) == 0:
        # No utilities found with this eia id
        console.print(
            f'Utility "{utility.name}" with EIA Id {utility.eia_id} not found in arcadia database', style="bold red"
        )
        return None
    if len(lses) == 1:
        # Found one utility
        utility_lse_id = lses[0]["lse_id"]
        return utility_lse_id
    else:
        # Nothing found; this should *theoretically* never happen but let's keep it just in case
        choices = [questionary.Choice(title=_["name"], value=_["lse_id"]) for _ in lses]
        choices.append(questionary.Separator())
        choices.append(questionary.Choice(title="None of these", value=None))
        utility_lse_id = cast(
            int | None,
            questionary.select(
                message=f"Found multiple utilities with lse id = {utility.eia_id}. Select one.", choices=choices
            ).ask(),
        )
        if utility_lse_id is None:
            console.print("No utility chosen")
            return None
        return utility_lse_id


def _select_tariffs(
    api: ArcadiaSignalAPI, lse_id: int, customer_classes: list[CustomerClass], tariff_types: list[TariffType], year: int
) -> list[tuple[str, int]]:
    with console.status("Fetching tariffs..."):
        tariffs = list(
            api.tariffs.iter_pages(
                lse_id=lse_id,
                effective_on=date(year, 6, 1),
                customer_classes=customer_classes,
                tariff_types=tariff_types,
            )
        )
    if not tariffs:
        return []
    return cast(
        list[tuple[str, int]],
        questionary.checkbox(
            message="Select tariffs",
            choices=[
                questionary.Choice(
                    title=f"{_['tariff_name']} ({_['tariff_id']})",
                    value=(_["tariff_name"], _["master_tariff_id"]),  # pyright: ignore[reportAny]
                    checked=True,
                )
                for _ in tariffs
            ],
            use_search_filter=True,
            use_jk_keys=False,
        ).ask(),
    )


def _select_customer_classes() -> list[CustomerClass]:
    return cast(
        list[CustomerClass],
        questionary.checkbox(
            message="Select customer classes",
            choices=[
                questionary.Choice(title="Residential", value="RESIDENTIAL"),
                questionary.Choice(title="General", value="GENERAL"),
                questionary.Choice(title="Special Use", value="SPECIAL_USE"),
            ],
            validate=lambda _: True if _ else "Select at least one customer class",
        ).ask(),
    )


def _select_tariff_types() -> list[TariffType]:
    return cast(
        list[TariffType],
        questionary.checkbox(
            message="Select tariff types",
            choices=[
                questionary.Choice(title="Default", value="DEFAULT"),
                questionary.Choice(title="Alternative", value="ALTERNATIVE"),
                questionary.Choice(title="Optional extra", value="OPTIONAL_EXTRA"),
                questionary.Choice(title="Rider", value="RIDER"),
            ],
            validate=lambda _: bool(_) or "Select at least one tariff type",
        ).ask(),
    )


def _fetch_tariffs(api: ArcadiaSignalAPI, tariffs: list[tuple[str, int]], year: int):
    result: list[TariffExtended] = []
    with console.status("Fetching tariffs..."):
        for name, id_ in tariffs:
            console.print(f"Tariff id: {name}")
            page = api.tariffs.iter_pages(
                fields="ext",
                master_tariff_id=id_,
                effective_on=date(year, 12, 1),
                populate_properties=True,
                populate_rates=True,
            )
            result.extend(page)
    return result


def _prompt_tariff_name(default: str) -> str:
    result = cast(str | None, questionary.text("Tariff name", default=default).ask())
    if result is None:
        exit()
    return result



import pytest

from tariff_fetch.urdb.rateacuity_history_gas.history_data import (
    DayOfMonth,
    RowValidationError,
    Season,
    _row_to_model,
)


def _base_row() -> dict[str, object]:
    row: dict[str, object] = {
        "rate": "Supply",
        "season": None,
        "effective_date": None,
        "start": None,
        "end": None,
        "determinant": None,
        "location": None,
        "rate_determinant": "per month",
    }
    for month in range(1, 13):
        row[f"{month:02d}/01/2025"] = 1.0
    return row


def test_row_to_model_parses_season_string() -> None:
    row = _base_row()
    row["season"] = "11/01 - 04/30"

    result = _row_to_model(row, location_avg_factor=1, month_column_names=list(row.keys())[-12:])

    assert result.season == Season(
        start=DayOfMonth(day=1, month=11),
        end=DayOfMonth(day=30, month=4),
    )


def test_row_to_model_rejects_invalid_season_string() -> None:
    row = _base_row()
    row["season"] = "winter"

    with pytest.raises(RowValidationError):
        _row_to_model(row, location_avg_factor=1, month_column_names=list(row.keys())[-12:])


@pytest.mark.parametrize("season", ["13/01 - 04/30", "00/15 - 01/01", "02/30 - 03/01", "04/31 - 05/01"])
def test_row_to_model_rejects_impossible_season_dates(season: str) -> None:
    row = _base_row()
    row["season"] = season

    with pytest.raises(RowValidationError):
        _row_to_model(row, location_avg_factor=1, month_column_names=list(row.keys())[-12:])


def test_row_to_model_allows_february_29_season_boundary() -> None:
    row = _base_row()
    row["season"] = "02/29 - 03/01"

    result = _row_to_model(row, location_avg_factor=1, month_column_names=list(row.keys())[-12:])

    assert result.season == Season(
        start=DayOfMonth(day=29, month=2),
        end=DayOfMonth(day=1, month=3),
    )



import json
import os
from datetime import UTC, datetime
from pathlib import Path
from typing import Literal, cast

import questionary
from dotenv import load_dotenv

from tariff_fetch.openei.utility_rates import UtilityRateSector, UtilityRatesResponseItem, iter_utility_rates

from . import console, prompt_filename
from .types import Utility


def _prompt_sector() -> UtilityRateSector:
    return cast(
        UtilityRateSector,
        questionary.select(
            message="Select sector",
            choices=[
                "Residential",
                "Commercial",
                "Industrial",
                "Lighting",
            ],
        ).ask(),
    )


def _prompt_detail_level() -> Literal["full", "minimal"]:
    return cast(
        Literal["full", "minimal"],
        questionary.select(
            message="Select level of detail",
            choices=["full", "minimal"],
        ).ask(),
    )


def _get_tariffs(
    eia_id: int, sector: UtilityRateSector, detail: Literal["full", "minimal"]
) -> list[UtilityRatesResponseItem]:
    api_key = os.getenv("OPENEI_API_KEY")
    if not api_key:
        raise ValueError("API Key is not set (via OPENEI_API_KEY variable)")
    with console.status("Fetching rates..."):
        iterator = iter_utility_rates(
            api_key,
            effective_on_date=datetime.now(UTC),
            sector=sector,
            detail=detail,
            eia=eia_id,
        )
        return list(iterator)


def _prompt_tariffs(tariffs: list[UtilityRatesResponseItem]) -> list[UtilityRatesResponseItem]:
    return cast(
        list[UtilityRatesResponseItem],
        questionary.checkbox(
            message="Select tariffs to include",
            choices=[questionary.Choice(title=_["name"], value=_, checked=True) for _ in tariffs],
        ).ask(),
    )


def process_openei(utility: Utility, output_folder: Path):
    _ = load_dotenv()
    if not os.getenv("OPENEI_API_KEY"):
        console.print("[b]OPENEI_API_KEY[/] environment variable is not set")
        console.print("Cannot use OpenEI API due to missing credentials")
        _ = console.input("Press enter to proceed...")
        return

    if not (sector := _prompt_sector()):
        return
    if not (detail_level := _prompt_detail_level()):
        return
    tariffs = _get_tariffs(utility.eia_id, sector, detail_level)
    if not tariffs:
        console.print("[red]No tariffs found[/]")
        _ = console.input("Press enter to proceed...")
        return
    tariffs = _prompt_tariffs(tariffs)
    if not tariffs:
        console.print("[red]No tariffs selected[/]")
        _ = console.input("Press enter to proceed...")
        return

    suggested_filename = f"openei_{utility.name}_{sector}_{detail_level}"
    if not (filepath := prompt_filename(output_folder, suggested_filename, "json")):
        return

    filepath.parent.mkdir(exist_ok=True)
    print(filepath)
    _ = filepath.write_text(json.dumps(tariffs, indent=2))
    console.print(f"Wrote [blue]{len(tariffs)}[/] items to {filepath}")



import json
from datetime import date, datetime
from pathlib import Path

from tariff_fetch.urdb.arcadia.library import LibraryDebugStore


def test_library_debug_store_saves_tariff_lookup_and_property(tmp_path: Path):
    store = LibraryDebugStore(tmp_path / "arcadia_library")

    store.save_tariff(
        {
            "tariff_id": 10,
            "master_tariff_id": 20,
            "effective_date": date(2025, 1, 1),
        }  # type: ignore[reportArgumentType]
    )
    store.save_lookups(
        "supplyCharge",
        2025,
        [
            {
                "from_date_time": datetime(2025, 1, 1, 0, 0),
                "to_date_time": datetime(2025, 2, 1, 0, 0),
                "actual_value": 0.1,
                "best_value": None,
                "forecast_value": None,
            }
        ],  # type: ignore[reportArgumentType]
    )
    store.save_property_value("territoryId", ["1", "2"])

    tariff_path = tmp_path / "arcadia_library" / "tariffs" / "master-20_tariff-10_effective-2025-01-01.json"
    lookups_path = tmp_path / "arcadia_library" / "lookups" / "supplyCharge_2025.json"
    property_path = tmp_path / "arcadia_library" / "properties" / "territoryId.json"

    assert json.loads(tariff_path.read_text())["tariff_id"] == 10
    assert json.loads(lookups_path.read_text())[0]["actual_value"] == 0.1
    assert json.loads(property_path.read_text()) == {"key": "territoryId", "value": ["1", "2"]}



"""Arcadia data access, caching, prompting, and debug persistence for conversion."""

import json
from datetime import date, datetime
from pathlib import Path
from typing import Literal, final, overload

from requests import HTTPError

from tariff_fetch.arcadia.api import ArcadiaSignalAPI
from tariff_fetch.arcadia.schema.lookup import Lookup
from tariff_fetch.arcadia.schema.tariff import TariffExtended
from tariff_fetch.arcadia.schema.tariffproperty import TariffPropertyPrunedDataType, TariffPropertyStandard
from tariff_fetch.arcadia.schema.tariffrate import TariffRateExtended
from tariff_fetch.urdb.arcadia.exception import TariffAccessDenied, TariffNotFoundByDate, TariffNotFoundById

from .exception import ConversionError
from .prompts import (
    prompt_boolean,
    prompt_choice,
    prompt_date,
    prompt_decimal,
    prompt_demand,
    prompt_integer,
    prompt_string,
)
from .shared import as_naive_datetime

PropertyValue = str | list[str] | bool | date | float | int
_DEFAULT_DEBUG_ROOT = Path("./outputs/arcadia_library")


def _json_default(value: object) -> str:
    """Serialize date-like values used in debug JSON dumps."""

    if isinstance(value, datetime | date):
        return value.isoformat()
    raise TypeError(f"Object of type {type(value).__name__} is not JSON serializable")


@final
class LibraryDebugStore:
    """Persist fetched Arcadia artifacts to disk for debugging."""

    def __init__(self, root: Path = _DEFAULT_DEBUG_ROOT) -> None:
        self.root = root
        self.tariffs_dir = root / "tariffs"
        self.lookups_dir = root / "lookups"
        self.properties_dir = root / "properties"

    def save_tariff(self, tariff: TariffExtended) -> None:
        """Write one fetched tariff version to the debug output folder."""

        effective_date = tariff["effective_date"].isoformat()
        tariff_id = tariff["tariff_id"]
        master_tariff_id = tariff["master_tariff_id"]
        path = self.tariffs_dir / f"master-{master_tariff_id}_tariff-{tariff_id}_effective-{effective_date}.json"
        self._write_json(path, tariff)

    def save_lookups(self, key: str, year: int, lookups: list[Lookup]) -> None:
        """Write one variable lookup timeseries to the debug output folder."""

        path = self.lookups_dir / f"{key}_{year}.json"
        self._write_json(path, lookups)

    def save_property_value(self, key: str, value: PropertyValue) -> None:
        """Write one prompted property value to the debug output folder."""

        path = self.properties_dir / f"{key}.json"
        self._write_json(path, {"key": key, "value": value})

    def _write_json(self, path: Path, payload: object) -> None:
        """Create the target directory and write formatted JSON payload."""

        path.parent.mkdir(parents=True, exist_ok=True)
        _ = path.write_text(json.dumps(payload, indent=2, default=_json_default))


@final
class TariffLibrary:
    """Fetch and cache Arcadia tariff versions and related tariff metadata."""

    def __init__(self, api: ArcadiaSignalAPI, debug_store: LibraryDebugStore) -> None:
        self.api = api
        self.debug_store = debug_store
        self.tariffs: list[TariffExtended] = []
        self._access_denied_tariff_ids: set[int] = set()

    def get_tariff_at_date(self, master_tariff_id: int, dt: date) -> TariffExtended:
        """Return the tariff version effective for a master tariff on a given date."""

        dt = dt.date() if isinstance(dt, datetime) else dt
        if (found := self._find_tariff_at_date(master_tariff_id, dt)) is not None:
            return found
        tariff = self._fetch_tariff_at_date(master_tariff_id, dt)
        self._remember(tariff)
        return tariff

    def get_tariff(self, tariff_id: int) -> TariffExtended:
        """Return a tariff by version-specific Arcadia tariff id."""

        if (found := self._find_tariff(tariff_id)) is not None:
            return found
        if tariff_id in self._access_denied_tariff_ids:
            raise TariffAccessDenied(tariff_id)
        tariff = self._fetch_tariff(tariff_id)
        self._remember(tariff)
        return tariff

    def get_rate(self, rate_id: int) -> TariffRateExtended:
        """Return a cached rate by Arcadia tariff rate id."""

        return next(
            rate for tariff in self.tariffs for rate in tariff.get("rates", []) if rate["tariff_rate_id"] == rate_id
        )

    def get_property(self, key: str) -> TariffPropertyStandard:
        """Return a cached Arcadia tariff property definition by key name."""

        return next(prop for tariff in self.tariffs for prop in tariff.get("properties", []) if prop["key_name"] == key)

    def _find_tariff(self, tariff_id: int) -> TariffExtended | None:
        """Search the in-memory tariff cache by tariff id."""

        return next((t for t in self.tariffs if t["tariff_id"] == tariff_id), None)

    def _find_tariff_at_date(self, master_tariff_id: int, dt: date) -> TariffExtended | None:
        """Search the in-memory cache for the tariff version effective on a given date."""

        for tariff in self.tariffs:
            if tariff["master_tariff_id"] != master_tariff_id:
                continue
            if _is_tariff_effective_on(tariff, dt):
                return tariff
        return None

    def _remember(self, tariff: TariffExtended) -> None:
        """Add a fetched tariff to the cache and write its debug dump."""

        self.tariffs.append(tariff)
        self.debug_store.save_tariff(tariff)

    def _fetch_tariff(self, tariff_id: int) -> TariffExtended:
        """Fetch one tariff version from Arcadia by tariff id."""

        try:
            tariffs = list(
                self.api.tariffs.iter_pages(
                    fields="ext",
                    search=str(tariff_id),
                    search_on=["tariffId"],
                    starts_with=True,
                    ends_with=True,
                    populate_properties=True,
                    populate_rates=True,
                )
            )
        except HTTPError as e:
            if e.response is not None and e.response.status_code == 403:
                self._access_denied_tariff_ids.add(tariff_id)
                raise TariffAccessDenied(tariff_id) from e
            raise
        if len(tariffs) > 1:
            raise RuntimeError(f"More than one tariff found for id={tariff_id}")
        if not tariffs:
            raise TariffNotFoundById(tariff_id)
        return tariffs[0]

    def _fetch_tariff_at_date(self, master_tariff_id: int, dt: date) -> TariffExtended:
        """Fetch the tariff version effective for a master tariff on a given date."""

        tariffs = self.api.tariffs.iter_pages(
            fields="ext",
            master_tariff_id=master_tariff_id,
            effective_on=(dt.date() if isinstance(dt, datetime) else dt),
            populate_properties=True,
            populate_rates=True,
        )
        try:
            return next(t for t in tariffs if _is_tariff_effective_on(t, dt))
        except StopIteration as e:
            raise TariffNotFoundByDate(master_tariff_id, dt) from e


@final
class VariablePropertyLibrary:
    """Fetch and cache Arcadia variable lookup timeseries by property key and year."""

    def __init__(self, api: ArcadiaSignalAPI, debug_store: LibraryDebugStore):
        self.api = api
        self.debug_store = debug_store
        self.property_timeseries: dict[tuple[str, int], list[Lookup]] = {}

    def lookup(self, key: str, dt: datetime) -> float:
        """Return the best available lookup value for a property at a datetime."""

        if (lookups := self.property_timeseries.get((key, dt.year))) is None:
            lookups = self._lookup_property_timeseries(key, dt.year)
            self.property_timeseries[(key, dt.year)] = lookups
            self.debug_store.save_lookups(key, dt.year, lookups)

        for row in lookups:
            if (
                as_naive_datetime(row["from_date_time"])
                <= as_naive_datetime(dt)
                <= as_naive_datetime(row["to_date_time"] or datetime.max)
            ):
                if (value := row["actual_value"]) is not None:
                    return value
                if (value := row["best_value"]) is not None:
                    return value
                if (value := row["forecast_value"]) is not None:
                    return value
                return 0.0
        return 0.0

    def _lookup_property_timeseries(self, key: str, year: int) -> list[Lookup]:
        """Fetch one year of lookup rows for a variable Arcadia property."""

        return list(
            self.api.properties.lookups.iter_pages(
                key,
                from_date_time=date(year, 1, 1),
                to_date_time=date(year + 1, 1, 1),
            )
        )


@final
class Library:
    """Top-level Arcadia conversion helper for tariffs, lookups, and prompted properties."""

    def __init__(
        self,
        api: ArcadiaSignalAPI,
        properties: dict[str, PropertyValue] | None = None,
        debug_root: Path = _DEFAULT_DEBUG_ROOT,
    ):
        self.api = api
        self.debug_store = LibraryDebugStore(debug_root)
        self.tariffs = TariffLibrary(api, self.debug_store)
        self.variables = VariablePropertyLibrary(api, self.debug_store)
        self._properies: dict[str, PropertyValue] = properties or {}
        self._issues: dict[tuple[object, ...], str] = {}

    def has_property(self, key: str) -> bool:
        """Return whether a property value has already been supplied or cached."""

        return key in self._properies

    def record_issue(self, key: tuple[object, ...], message: str) -> None:
        """Record a deduplicated conversion issue for this conversion run."""

        _ = self._issues.setdefault(key, message)

    def iter_issues(self) -> list[str]:
        """Return the unique conversion issues recorded during this conversion run."""

        return list(self._issues.values())

    def get_choice_property_as_ints(self, key: str) -> list[int]:
        """Return a CHOICE property as integer ids."""

        strs = self.get_property(key, "CHOICE")
        try:
            return list(map(int, strs))
        except ValueError as e:
            raise ConversionError(f"Could not convert value of choice property {key} to strings") from e

    @overload
    def get_property(self, key: str, data_type: Literal["STRING"]) -> str: ...

    @overload
    def get_property(self, key: str, data_type: Literal["CHOICE"]) -> list[str]: ...

    @overload
    def get_property(self, key: str, data_type: Literal["BOOLEAN"]) -> bool: ...
    @overload
    def get_property(self, key: str, data_type: Literal["DATE"]) -> date: ...
    @overload
    def get_property(self, key: str, data_type: Literal["DECIMAL"]) -> float: ...
    @overload
    def get_property(self, key: str, data_type: Literal["INTEGER"]) -> int: ...
    @overload
    def get_property(self, key: str, data_type: Literal["DEMAND"]) -> float: ...

    def get_property(self, key: str, data_type: TariffPropertyPrunedDataType) -> PropertyValue:
        """Return a property value, prompting and caching it if needed."""

        if (found := self._get_property(key, data_type)) is not None:
            return found
        tariff_property = self.tariffs.get_property(key)
        result = _prompt_property(tariff_property)
        if result is None:
            raise ConversionError("Property not set")
        self._properies[key] = result
        self.debug_store.save_property_value(key, result)
        return result

    @overload
    def _get_property(self, key: str, data_type: Literal["STRING"]) -> str | None: ...

    @overload
    def _get_property(self, key: str, data_type: Literal["CHOICE"]) -> list[str] | None: ...

    @overload
    def _get_property(self, key: str, data_type: Literal["BOOLEAN"]) -> bool | None: ...
    @overload
    def _get_property(self, key: str, data_type: Literal["DATE"]) -> date | None: ...
    @overload
    def _get_property(self, key: str, data_type: Literal["DECIMAL"]) -> float | None: ...
    @overload
    def _get_property(self, key: str, data_type: Literal["INTEGER"]) -> int | None: ...
    @overload
    def _get_property(self, key: str, data_type: Literal["DEMAND"]) -> float | None: ...

    def _get_property(self, key: str, data_type: TariffPropertyPrunedDataType) -> PropertyValue | None:
        """Return a cached property value after validating its expected data type."""

        try:
            value = self._properies[key]
        except KeyError:
            return None
        value_type = type(value)
        match data_type:
            case "STRING":
                if not isinstance(value, str):
                    raise ConversionError(f"Value for property {key} is expected to be `str`, not {value_type}")
                return value
            case "CHOICE":
                if not isinstance(value, list):
                    raise ConversionError(f"Value for property {key} is expected to be a list, not {value_type}")
                return value
            case "BOOLEAN":
                if not isinstance(value, bool):
                    raise ConversionError(f"Value for property {key} is expected to be `bool`, not {value_type}")
                return value
            case "DATE":
                if not isinstance(value, date):
                    raise ConversionError(f"Value for property {key} is expected to be `bool`, not {value_type}")
                return value
            case "DECIMAL":
                if not isinstance(value, float):
                    raise ConversionError(f"Value for property {key} is expected to be `float`, not {value_type}")
                return value
            case "INTEGER":
                if not isinstance(value, int):
                    raise ConversionError(f"Value for property {key} is expected to be `int`, not {value_type}")
                return value

            case "DEMAND":
                if not isinstance(value, float):
                    raise ConversionError(f"Value for property {key} is expected to be `float`, not {value_type}")
                return value


def _is_tariff_effective_on(tariff: TariffExtended, dt: date) -> bool:
    """Return whether a tariff version is effective on the given date."""

    effective_date = tariff["effective_date"]
    end_date = tariff["end_date"] or date.max
    return effective_date <= dt < end_date


def _prompt_property(tariff_property: TariffPropertyStandard) -> PropertyValue | None:
    """Prompt the user for a property value using the Arcadia property data type."""

    data_type = tariff_property["data_type"]
    match data_type:
        case "BOOLEAN":
            return prompt_boolean(tariff_property)
        case "CHOICE":
            return prompt_choice(tariff_property)
        case "STRING":
            return prompt_string(tariff_property)
        case "DATE":
            return prompt_date(tariff_property)
        case "DECIMAL":
            return prompt_decimal(tariff_property)
        case "INTEGER":
            return prompt_integer(tariff_property)
        case "DEMAND":
            return prompt_demand(tariff_property)
        case "FORMULA":
            raise ValueError("Formula properties are not supported")



import contextlib
import re
from calendar import monthrange
from collections.abc import Iterator
from datetime import datetime
from math import inf
from typing import Any, NamedTuple, cast, final

import polars as pl
from pydantic import BaseModel, TypeAdapter, ValidationError, field_validator
from typing_extensions import override

from .exceptions import IncorrectDataframeSchemaMonths, IncorrectDataframeSchemaMultipleYears
from .shared import is_date_column_name, kwh_multiplier
from .types import (
    BandDeterminant,
    ConsumptionRateDeterminant,
    FixedRateDeterminant,
    PercentRateDeterminant,
)


class DayOfMonth(NamedTuple):
    day: int
    month: int


class Season(NamedTuple):
    start: DayOfMonth
    end: DayOfMonth


_SEASON_PATTERN = re.compile(r"^\s*(\d{2})/(\d{2})\s*-\s*(\d{2})/(\d{2})\s*$")


@final
class RowValidationError(Exception):
    def __init__(self, row: dict[str, Any]) -> None:  # pyright: ignore[reportExplicitAny]
        super().__init__()
        self.row = row

    @override
    def __str__(self) -> str:
        return f"Cannot validate row: {self.row}"


@final
class HistoryData:
    def __init__(self, df: pl.DataFrame) -> None:
        self._df = df

    def rows(self) -> Iterator["Row"]:
        month_column_names = _get_month_column_names(self._df)
        location_avg_factor = self.location_avg_factor()
        for row_dict in self._df.iter_rows(named=True):
            with contextlib.suppress(RowValidationError):
                yield _row_to_model(row_dict, location_avg_factor, month_column_names)

    def get_unknown_nonempty_columns(self) -> list[str]:
        df = self._df
        df_schema = df.schema
        non_date_columns = {col for col in df_schema if not is_date_column_name(col)}
        allow_empty_columns = non_date_columns - FixedChargeRow.model_fields.keys()
        return [
            c
            for c in allow_empty_columns
            if not (df[c].is_null() | (df[c] == "" if df[c].dtype == pl.Utf8 else False)).all()
        ]

    def validate_rows(self) -> list[RowValidationError]:
        result: list[RowValidationError] = []
        month_column_names = _get_month_column_names(self._df)
        location_avg_factor = self.location_avg_factor()
        for row_dict in self._df.iter_rows(named=True):
            try:
                _ = _row_to_model(row_dict, location_avg_factor, month_column_names)
            except RowValidationError as e:
                result.append(e)
        return result

    def location_avg_factor(self) -> float:
        if "location" not in self._df.schema:
            return 1
        count = cast(
            int,
            self._df.select(  # pyright: ignore[reportUnknownMemberType]
                pl.col("location").filter(pl.col("location").is_not_null() & (pl.col("location") != "")).n_unique()
            ).item(),
        )
        if count == 0:
            return 1
        return 1 / count


class _Row(BaseModel):
    rate: str
    season: Season | None
    year: int
    effective_date: str | None
    start: float | None = None
    end: float | None = None
    # rate_determinant: RateDeterminant
    determinant: BandDeterminant | None = None
    location: str | None = None
    month_values: list[float | None]
    location_avg_factor: float

    @field_validator("season", mode="before")
    @classmethod
    def _parse_season(cls, value: object) -> object:
        if value is None or value == "":
            return None
        if isinstance(value, Season):
            return value
        if not isinstance(value, str):
            return value
        match = _SEASON_PATTERN.fullmatch(value)
        if match is None:
            return value
        start_month, start_day, end_month, end_day = (int(part) for part in match.groups())
        return Season(
            start=_validated_day_of_month(start_month, start_day),
            end=_validated_day_of_month(end_month, end_day),
        )

    @property
    def start_kwh(self) -> float:
        if self.determinant is None:
            return 0
        if self.start is not None:
            return round(self.start * kwh_multiplier(self.determinant))
        return 0

    @property
    def end_kwh(self) -> float:
        if self.determinant is None:
            return inf
        if self.end is not None:
            return round(self.end * kwh_multiplier(self.determinant))
        return inf


class ConsumptionRow(_Row):
    rate_determinant: ConsumptionRateDeterminant

    def month_value_kwh(
        self,
        month: int,
    ) -> float:
        raw_value = self.month_values[month]
        if raw_value is None:
            return 0
        result = raw_value / kwh_multiplier(self.rate_determinant)
        if self.location:
            result *= self.location_avg_factor
        return result


class PercentageRow(_Row):
    rate_determinant: PercentRateDeterminant

    def month_value_float(self, month: int) -> float:
        raw_value = self.month_values[month]
        if raw_value is None:
            return 0
        result = raw_value / 100
        if self.location:
            result *= self.location_avg_factor
        return result


class FixedChargeRow(_Row):
    rate_determinant: FixedRateDeterminant

    def month_value(self, month: int) -> float:
        raw_value = self.month_values[month]
        if raw_value is None:
            return 0
        return raw_value


Row = ConsumptionRow | PercentageRow | FixedChargeRow


def _row_to_model(row: dict[str, Any], location_avg_factor: float, month_column_names: list[str]) -> Row:  # pyright: ignore[reportExplicitAny]
    result = row.copy()
    result["month_values"] = []
    result["year"] = datetime.strptime(month_column_names[0], "%m/%d/%Y").year
    for col in month_column_names:
        value = cast(float | None, row[col])
        del result[col]
        result["month_values"].append(value)  # pyright: ignore[reportUnknownMemberType]
    result["location_avg_factor"] = location_avg_factor
    ta: TypeAdapter[Row] = TypeAdapter(Row)
    try:
        return ta.validate_python(result)
    except ValidationError:
        raise RowValidationError(result) from None


def _get_month_column_names(df: pl.DataFrame):
    date_columns = [col for col in df.columns if is_date_column_name(col)]
    date_columns = sorted(date_columns, key=lambda c: datetime.strptime(c, "%m/%d/%Y").month)
    date_columns_datetimes = [datetime.strptime(c, "%m/%d/%Y") for c in date_columns]
    if [c.month for c in date_columns_datetimes] != list(range(1, 13)):
        raise IncorrectDataframeSchemaMonths()
    if len({c.year for c in date_columns_datetimes}) != 1:
        raise IncorrectDataframeSchemaMultipleYears()
    return date_columns


def _validated_day_of_month(month: int, day: int) -> DayOfMonth:
    if not 1 <= month <= 12:
        raise ValueError(f"Invalid season month: {month}")
    # Use a leap year so recurring seasonal boundaries can represent Feb 29.
    if not 1 <= day <= monthrange(2024, month)[1]:
        raise ValueError(f"Invalid season day {day} for month {month}")
    return DayOfMonth(day=day, month=month)



import logging
from datetime import date
from pathlib import Path
from typing import Annotated, cast

import polars as pl
import questionary
import typer
from requests import HTTPError
from rich.prompt import Prompt

from tariff_fetch._cli.arcadia_urdb import process_genability as process_genability_urdb
from tariff_fetch._cli.genability import process_genability
from tariff_fetch._cli.openei import process_openei
from tariff_fetch._cli.rateacuity import process_rateacuity
from tariff_fetch.rateacuity.base import AuthorizationError

from ._cli import console
from ._cli.types import Provider, StateCode, Utility

ENTITY_TYPES_SORTORDER = ["Investor Owned", "Cooperative", "Municipal"]
CORE_EIA861_YEARLY_SALES_HTTPS = (
    "https://s3.us-west-2.amazonaws.com/pudl.catalyst.coop/nightly/core_eia861__yearly_sales.parquet"
)


def prompt_year() -> int:
    result = cast(
        str, questionary.text("Enter year", default=str(date.today().year - 1), validate=_is_valid_year).ask()
    )
    return int(result)


def _is_valid_year(value: str) -> bool:
    try:
        _ = date(int(value), 1, 1)
    except (TypeError, ValueError):
        return False
    return True


def prompt_state() -> StateCode:
    choice = Prompt.ask(
        "Enter two-letter state abbreviation",
        choices=[state.value for state in StateCode],
        show_choices=False,
        case_sensitive=False,
    )
    return StateCode(choice.lower())


def prompt_providers() -> list[Provider]:
    return cast(
        list[Provider],
        questionary.checkbox(
            message="Select providers",
            choices=[questionary.Choice(title=_.value, value=_) for _ in Provider],
            validate=lambda x: True if x else "Select at least one provider",
        ).ask(),
    )


def prompt_utility(state: str) -> Utility:
    with console.status("Fetching utilities..."):
        yearly_sales_df = (
            pl.read_parquet(CORE_EIA861_YEARLY_SALES_HTTPS)  # pyright: ignore[reportUnknownMemberType]
            .filter(pl.col("state") == state.upper())
            .filter(pl.col("report_date") == pl.col("report_date").max().over("utility_id_eia"))  # pyright: ignore[reportUnknownMemberType]
            .filter(pl.col("entity_type").is_in(ENTITY_TYPES_SORTORDER))
            .group_by("utility_id_eia")
            .agg(
                pl.col("utility_name_eia").last().alias("utility_name"),
                pl.col("business_model").last().alias("business_model"),
                pl.col("sales_mwh").filter(pl.col("customer_class") == "residential").sum().alias("sales_mwh"),
                pl.col("sales_revenue").sum().alias("sales_revenue"),
                pl.col("customers").filter(pl.col("customer_class") == "residential").sum().alias("customers"),
                pl.col("entity_type").last().alias("entity_type"),
            )
            .sort(["entity_type", "customers", "utility_name"], descending=(False, True, False))
        )

        rows = list(yearly_sales_df.iter_rows(named=True))
        rows.sort(
            key=lambda _: (
                ENTITY_TYPES_SORTORDER.index(_["entity_type"])  # pyright: ignore[reportAny]
                if _["entity_type"] in ENTITY_TYPES_SORTORDER
                else abs(hash(_["entity_type"])) + 4,  # pyright: ignore[reportAny]
                -_["customers"],
                _["utility_name"],
            )
        )

    def fmt_number(value: float | int | None) -> str:
        if value is None:
            return "-"
        return f"{value:,.0f}"

    utility_name_header = "Utility Name"
    entity_type_header = "Entity Type"
    sales_header = "Sales (MWh)"
    revenue_header = "Revenue ($)"
    customers_header = "Customers"

    largest_utility_name = max(len(utility_name_header), *(len(row["utility_name"]) for row in rows))  # pyright: ignore[reportAny]
    largest_entity_type = max(len(entity_type_header), *(len(row["entity_type"][:18]) for row in rows))  # pyright: ignore[reportAny]
    largest_sales_col = max(len(sales_header), *(len(fmt_number(row["sales_mwh"])) for row in rows))  # pyright: ignore[reportAny]
    largest_revenue_col = max(len(revenue_header), *(len(fmt_number(row["sales_revenue"])) for row in rows))  # pyright: ignore[reportAny]
    largest_customers_col = max(len(customers_header), *(len(fmt_number(row["customers"])) for row in rows))  # pyright: ignore[reportAny]

    header_str_utility_name = utility_name_header.ljust(largest_utility_name)
    header_str_entity_type = entity_type_header.ljust(largest_entity_type)
    header_str_sales = sales_header.ljust(largest_sales_col)
    header_str_revenue = revenue_header.ljust(largest_revenue_col)
    header_str_customers = customers_header.ljust(largest_customers_col)
    header_str = f"{header_str_utility_name} | {header_str_entity_type} | {header_str_sales} | {header_str_revenue} | {header_str_customers}"
    separator = questionary.Separator(line="-" * len(header_str))

    header = questionary.Choice(
        title=header_str,
        value=0,
    )

    def build_choice(row: dict[str, str | int | float | None]) -> questionary.Choice:
        name_col = cast(str, row["utility_name"]).ljust(largest_utility_name)
        entity_type = (cast(str, row["entity_type"]) or "-")[:18].ljust(largest_entity_type)
        sales_col = fmt_number(cast(float, row["sales_mwh"])).ljust(largest_sales_col)
        revenue_col = fmt_number(cast(float, row["sales_revenue"])).ljust(largest_revenue_col)
        customers_col = fmt_number(cast(float, row["customers"])).ljust(largest_customers_col)
        title = f"{name_col} | {entity_type} | {sales_col} | {revenue_col} | {customers_col}"
        return questionary.Choice(
            title=title,
            value=Utility(eia_id=cast(int, row["utility_id_eia"]), name=cast(str, row["utility_name"])),
        )

    result = 0
    while result == 0:
        result = cast(
            Utility,
            questionary.select(
                message="Select a utility",
                choices=[header, separator, *[build_choice(row) for row in rows]],
                use_search_filter=True,
                use_jk_keys=False,
                use_shortcuts=False,
            ).ask(),
        )
    return result


def main(
    state: Annotated[
        StateCode | None, typer.Option("--state", "-s", help="Two-letter state abbreviation", case_sensitive=False)
    ] = None,
    providers: Annotated[list[Provider] | None, typer.Option("--providers", "-p", case_sensitive=False)] = None,
    output_folder: Annotated[
        str, typer.Option("--output-folder", "-o", help="Folder to store outputs in")
    ] = "./outputs",
    urdb: bool = False,
):
    logging.basicConfig(level=logging.DEBUG)
    # print(pl.read_parquet(CoreEIA861_ASSN_UTILITY.https))
    state_ = state or prompt_state().value
    providers_ = providers or prompt_providers()
    output_folder_ = Path(output_folder)
    utility = prompt_utility(state_)

    if urdb:
        year = prompt_year()
        if Provider.GENABILITY in providers_:
            console.print("Processing [blue]Genability[/]")
            try:
                process_genability_urdb(utility=utility, output_folder=output_folder_, year=year)
            except HTTPError as e:
                if e.response.status_code == 401:
                    console.print("Authorization failed")
                    console.print(
                        "Check if credentials set via [b]ARCADIA_APP_ID[/] and [b]ARCADIA_APP_KEY[/] environment variables are correct"
                    )
                else:
                    raise e from None
        if Provider.OPENEI in providers_:
            console.print("Processing [blue]OpenEI[/]")
            try:
                process_openei(utility, output_folder_)
            except HTTPError as e:
                if e.response.status_code == 403:
                    console.print("Authorization failed")
                    console.print("Check if [b]OPENEI_API_KEY[/] environment variable is correct")
                else:
                    raise e from None
        if Provider.RATEACUITY in providers_:
            console.print("Cannot convert RateAcuity data to URDB")

    else:
        if Provider.GENABILITY in providers_:
            console.print("Processing [blue]Genability[/]")
            try:
                process_genability(utility=utility, output_folder=output_folder_)
            except HTTPError as e:
                if e.response.status_code == 401:
                    console.print("Authorization failed")
                    console.print(
                        "Check if credentials set via [b]ARCADIA_APP_ID[/] and [b]ARCADIA_APP_KEY[/] environment variables are correct"
                    )
                else:
                    raise e from None
        if Provider.OPENEI in providers_:
            console.print("Processing [blue]OpenEI[/]")
            try:
                process_openei(utility, output_folder_)
            except HTTPError as e:
                if e.response.status_code == 403:
                    console.print("Authorization failed")
                    console.print("Check if [b]OPENEI_API_KEY[/] environment variable is correct")
                else:
                    raise e from None
        if Provider.RATEACUITY in providers_:
            console.print("Processing [blue]RateAcuity[/]")
            try:
                process_rateacuity(output_folder_, state_, utility)
            except AuthorizationError:
                console.print("Authorization failed")
                console.print(
                    "Check if credentials provided via [b]RATEACUITY_USERNAME[/] and [b]RATEACUITY_PASSWORD[/] environment variables are correct"
                )


def main_cli():
    typer.run(main)


if __name__ == "__main__":
    main_cli()



from datetime import datetime
from typing import NotRequired

from pydantic.alias_generators import to_camel
from typing_extensions import TypedDict


class Lookup(TypedDict):
    __pydantic_config__ = {"alias_generator": to_camel, "extra": "allow"}  # pyright: ignore[reportGeneralTypeIssues, reportUnannotatedClassAttribute]  # noqa
    lookup_id: int
    """The unique ID for this lookup value entry in the lookup table"""
    property_key: str
    """The unique name for the property this lookup belongs to"""
    sub_property_key: NotRequired[str]
    """Sub-property this lookup value applies to (when needed, e.g., wholesale index node or zone)"""
    from_date_time: datetime
    to_date_time: datetime | None
    best_value: float | None
    """Best kWh/kW price for the property key for the date range"""
    best_accuracy: float | None
    """Reserved for future use"""
    actual_value: float | None
    """Actual kWh/kW price for the property key for the date range"""
    """kWh/kW price forecasted by the utility for the property key for the date range"""
    lse_forecast_value: float | None
    lse_forecast_accuracy: float | None
    """Reserved for future use"""
    forecast_value: float | None
    """kWh/kW price forecasted by Arcadia for the property key for the date range based on previous values"""
    forecast_accuracy: float | None
    """Reserved for future use"""



import datetime
from typing import Any
from urllib.parse import urljoin

import requests

BASE_URL = "https://api.openei.org"


def convert_params(input: dict[str, Any]) -> dict[str, Any]:  # pyright: ignore[reportExplicitAny]
    return {k: _convert_param(v) for k, v in input.items()}  # pyright: ignore[reportAny]


def _convert_param(value: Any):  # pyright: ignore[reportAny, reportExplicitAny]
    if isinstance(value, datetime.datetime):
        return int(value.timestamp())
    return value  # pyright:ignore[reportAny]
    # return {True: "true", False: "false"}.get(value, value)


def api_request_json(path: str, api_key: str, **params) -> dict | list:  # pyright: ignore[reportMissingTypeArgument, reportUnknownParameterType, reportMissingParameterType]
    params_ = {"api_key": api_key, **convert_params(params)}  # pyright: ignore[reportUnknownArgumentType]
    url = urljoin(BASE_URL, path)
    response = requests.get(url, params=params_)
    response.raise_for_status()
    return response.json()  # pyright: ignore[reportAny]



"""Shared type aliases used by the Arcadia-to-URDB converter."""

from collections.abc import Callable
from datetime import datetime

from tariff_fetch.arcadia.schema.common import RateChargeClass

ConsumptionBand = tuple[float, float]  # (upper_limit, value)
ConsumptionBandSet = list[ConsumptionBand]
DayPredicate = Callable[[datetime], bool]
PercentageModifiers = list[tuple[set[RateChargeClass], float]]



from calendar import Month
from datetime import date
import re
from typing import Annotated, Any, Generic, NamedTuple, Self, TypeVar, Literal

from typing_extensions import TypedDict

from pydantic import BeforeValidator
from pydantic.alias_generators import to_camel


USStateCode = Literal[
    "AL",
    "AK",
    "AZ",
    "AR",
    "CA",
    "CO",
    "CT",
    "DE",
    "FL",
    "GA",
    "HI",
    "ID",
    "IL",
    "IN",
    "IA",
    "KS",
    "KY",
    "LA",
    "ME",
    "MD",
    "MA",
    "MI",
    "MN",
    "MS",
    "MO",
    "MT",
    "NE",
    "NV",
    "NH",
    "NJ",
    "NM",
    "NY",
    "NC",
    "ND",
    "OH",
    "OK",
    "OR",
    "PA",
    "RI",
    "SC",
    "SD",
    "TN",
    "TX",
    "UT",
    "VT",
    "VA",
    "WA",
    "WV",
    "WI",
    "WY",
]

UseType = Literal["C", "R", "I", "A", " "]
"""
Indicator of what type of customer accounts the schedule can be used for.
C=Commercial, R=Residential, I=Industrial, A=Agricultural. Will be blank if used for
more than one type of account.
"""

SCRateDeterminant = Literal["per month"]
OCRateDeterminant = Literal["per month"]
CORateDeterminant = Literal["per therm"]


class MonthAndDay(NamedTuple):
    month: int
    day: int


def month_and_day(input_: str | MonthAndDay | None) -> MonthAndDay | None:
    if input_ is None:
        return None
    if isinstance(input_, tuple):
        return input_
    daystr, monthstr = input_.split("-")
    return MonthAndDay(
        month=("JAN", "FEB", "MAR", "APR", "MAY", "JUN", "JUL", "AUG", "SEP", "OCT", "NOV", "DEC").index(
            monthstr.upper()
        ),
        day=int(daystr),
    )


def empty_is_none(input_: Any) -> str | None:  # pyright: ignore[reportAny, reportExplicitAny]
    if not isinstance(input_, str):
        return input_  # pyright: ignore[reportAny]
    if not input_:
        return None
    return input_


empty_or_none_validator = BeforeValidator(empty_is_none)
FloatOrNone = Annotated[float | None, empty_or_none_validator]
DateOrNone = Annotated[date | None, empty_or_none_validator]
StrOrNone = Annotated[str | None, empty_or_none_validator]

MonthAndDayAnnotated = Annotated[MonthAndDay, BeforeValidator(month_and_day)]
MonthAndDayOrNone = Annotated[MonthAndDay | None, BeforeValidator(month_and_day), empty_or_none_validator]


def to_spaced_camel(snake_case: str) -> str:
    camel = to_camel(snake_case)  # snake_case -> lowerCamelCase
    spaced = re.sub(r"([A-Z])", r" \1", camel)
    return spaced.strip().title()


class Schedule(TypedDict):
    __pydantic_config__ = {"alias_generator": to_spaced_camel, "extra": "forbid"}  # pyright: ignore[reportGeneralTypeIssues, reportUnannotatedClassAttribute]  # noqa
    state: USStateCode
    utility: str
    schedule_name: str
    schedule_description: str
    use_type: UseType
    min_peak: FloatOrNone
    """ Start of specific peak range for which the schedule should be used """
    max_peak: FloatOrNone
    """End of specific peak range for which the schedule should be used"""
    peak_determinant: str
    """Used in conjunction with MinPeak and MaxPeak to define the unit of measure for peak values"""
    min_usage: FloatOrNone
    """Indicator of minimum usage of account served by the schedule"""
    max_usage: FloatOrNone
    """Indicator of maximum usage of account served by the schedule"""
    usage_determinant: str
    """Used in conjunction with MinUsage and MaxUsage to define the min and max measurement"""
    option_description: str
    """Text description of specific option from tariff for this data set"""
    load_season: str
    """
    If the schedule is applicable based on load during a specific season the season is shown here
    """
    load_comp: str
    """
    If the schedule is applicable on load during a specific season compared to
    other months, the months to compare are shown here
    """
    load_percent: FloatOrNone
    """
    If the schedule is applicable on load during a specific season compared to
    other months, the comparison value is shown here
    """
    season_start: DateOrNone
    """First date of season"""
    season_end: DateOrNone
    """Last date of season"""
    effective_date: date


class ServiceCharge(TypedDict):
    __pydantic_config__ = {"alias_generator": to_spaced_camel, "extra": "forbid"}  # pyright: ignore[reportGeneralTypeIssues, reportUnannotatedClassAttribute]  # noqa
    service_charges: str
    season_start: DateOrNone
    season_end: DateOrNone
    start: FloatOrNone
    """The minimum value for which this record should be used"""
    end: FloatOrNone
    """The maximum value for which this record should be used"""
    charge_determinant: str
    """Used with Start and End to show the measurement of the values of
    these fields"""
    rate: float
    rate_determinant: SCRateDeterminant
    """Measure of rate charge"""
    effective_date: date


class OtherCharge(TypedDict):
    __pydantic_config__ = {"alias_generator": to_spaced_camel, "extra": "forbid"}  # pyright: ignore[reportGeneralTypeIssues, reportUnannotatedClassAttribute]  # noqa
    other_charges: str
    rate: float
    rate_determinant: OCRateDeterminant
    min_therms: FloatOrNone
    max_therms: FloatOrNone
    effective_date: date


class Consumption(TypedDict):
    __pydantic_config__ = {"alias_generator": to_spaced_camel, "extra": "forbid"}  # pyright: ignore[reportGeneralTypeIssues, reportUnannotatedClassAttribute]  # noqa
    consumption: str
    season_start: MonthAndDayOrNone
    season_end: MonthAndDayOrNone
    rate: float
    rate_determinant: CORateDeterminant
    effective_date: date


T = TypeVar("T")


class Table(TypedDict, Generic[T]):
    title: str
    values: list[T]


ScheduleTable = Table[Schedule]
ServiceChargesTable = Table[ServiceCharge]
OtherChargesTable = Table[OtherCharge]
ConsumptionTable = Table[Consumption]

TableUnion = ScheduleTable | ServiceChargesTable | OtherChargesTable | ConsumptionTable


class Section(TypedDict):
    section: str | None
    tables: list[TableUnion]


class Tariff(TypedDict):
    schedule: str
    sections: list[Section]



"""Convert simple Arcadia flat demand charges into URDB flat demand fields."""

from datetime import datetime
from math import inf

from tariff_fetch.arcadia.schema.tariffrate import TariffRateBand, TariffRateExtended
from tariff_fetch.urdb.schema import FlatDemandTier, URDBRate

from . import rateutils as ru
from .exception import RateConversionError
from .library import Library
from .scenario import Scenario

_SUPPORTED_DEMAND_UNIT = "kW"


def build_flat_demand(scenario: Scenario, library: Library) -> URDBRate:
    """Build URDB flat-demand fields for a scenario year.

    This first implementation intentionally supports only simple monthly flat
    demand charges:

    - Arcadia `DEMAND_BASED` rates
    - monthly charge period
    - no TOU structure
    - no variable factors, variable limits, or property-driven formulas
    - BUY transaction type only
    - rate bands expressed as demand tiers
    """

    monthly_band_sets = [get_month_flat_demand_bands(scenario, library, month) for month in range(1, 13)]

    if all(not bands for bands in monthly_band_sets):
        return {}

    band_sets: list[tuple[FlatDemandTier, ...]] = []
    band_index: dict[tuple[FlatDemandTier, ...], int] = {}
    month_indexes: list[int] = []

    for bands in monthly_band_sets:
        key = tuple(bands)
        if key not in band_index:
            band_index[key] = len(band_sets)
            band_sets.append(key)
        month_indexes.append(band_index[key])

    return {
        "flatdemandunit": _SUPPORTED_DEMAND_UNIT,
        "flatdemandstructure": [list(bands) for bands in band_sets],
        "flatdemandmonths": tuple(month_indexes),
    }


def get_month_flat_demand_bands(scenario: Scenario, library: Library, month: int) -> list[FlatDemandTier]:
    """Return the combined flat-demand bands applicable for one month."""

    sample_dt = datetime(scenario.year, month, 15, 12)
    tariff = library.tariffs.get_tariff_at_date(scenario.master_tariff_id, sample_dt.date())
    rates = [
        rate
        for rate in tariff.get("rates", [])
        if ru.rate_is_applied_to_scenario(rate, scenario, library) and ru.rate_is_applied_to_datetime(rate, sample_dt)
    ]

    demand_bands: list[list[FlatDemandTier]] = []
    for rate in rates:
        bands = get_rate_flat_demand_bands(rate, scenario, library)
        if bands is None:
            continue
        demand_bands.append(bands)

    if not demand_bands:
        return []
    if len(demand_bands) == 1:
        return demand_bands[0]
    return _sum_flat_demand_band_sets(demand_bands)


def get_rate_flat_demand_bands(
    rate: TariffRateExtended, scenario: Scenario, library: Library
) -> list[FlatDemandTier] | None:
    """Convert one Arcadia demand-based rate into URDB flat-demand tiers."""

    if rate.get("charge_type") != "DEMAND_BASED":
        return None
    if rate.get("time_of_use") is not None:
        raise RateConversionError(rate, "TOU demand charges are not supported by flat demand conversion")
    if rate.get("quantity_key") is not None:
        raise RateConversionError(rate, "Rates with quantity_key are not supported for flat demand conversion")
    if rate["charge_period"] != "MONTHLY":
        raise RateConversionError(rate, "Flat demand conversion only supports MONTHLY charge period")
    if rate["transaction_type"] != "BUY":
        raise RateConversionError(rate, "Flat demand conversion only supports BUY transaction type")

    bands = [
        _validate_flat_demand_band(rate, band)
        for band in ru.rate_filter_bands(rate, scenario, library)
    ]
    return sorted(bands, key=lambda band: band.get("max", inf))


def _validate_flat_demand_band(rate: TariffRateExtended, band: TariffRateBand) -> FlatDemandTier:
    """Validate and convert one Arcadia demand band into a URDB flat-demand tier."""

    if band["rate_unit"] != "COST_PER_UNIT":
        raise RateConversionError(rate, "Flat demand bands must have rate unit = COST_PER_UNIT")
    if band["is_credit"]:
        raise RateConversionError(rate, "Credit demand rates are not supported")
    if band["has_consumption_limit"]:
        raise RateConversionError(rate, "Flat demand bands cannot have consumption limits")
    if band.get("consumption_upper_limit") is not None:
        raise RateConversionError(rate, "Flat demand bands cannot have consumption_upper_limit")
    if band["has_property_limit"]:
        raise RateConversionError(rate, "Flat demand bands cannot have property limits")
    if band.get("property_upper_limit") is not None:
        raise RateConversionError(rate, "Flat demand bands cannot have property_upper_limit")
    if band.get("prev_upper_limit") not in {None, 0.0}:
        raise RateConversionError(rate, "Flat demand bands cannot have prev_upper_limit")

    tier: FlatDemandTier = {"rate": band["rate_amount"]}
    if band["has_demand_limit"] or band.get("demand_upper_limit") is not None:
        if (limit := band.get("demand_upper_limit")) is None:
            raise RateConversionError(rate, "Flat demand band has_demand_limit=true but no demand_upper_limit")
        tier["max"] = limit
    return tier


def _sum_flat_demand_band_sets(inputs: list[list[FlatDemandTier]]) -> list[FlatDemandTier]:
    """Sum multiple flat-demand tier sets over their combined demand breakpoints."""

    limits = sorted({tier["max"] for tiers in inputs for tier in tiers if "max" in tier})
    if any("max" not in tiers[-1] for tiers in inputs):
        limits.append(inf)

    def value_at(tiers: list[FlatDemandTier], limit: float) -> float:
        for tier in tiers:
            if limit <= tier.get("max", inf):
                return tier["rate"]
        return 0.0

    result: list[FlatDemandTier] = []
    for limit in limits:
        rate = sum(value_at(tiers, limit) for tiers in inputs)
        tier: FlatDemandTier = {"rate": rate}
        if limit != inf:
            tier["max"] = limit
        result.append(tier)

    collapsed = [result[0]]
    for tier in result[1:]:
        if tier["rate"] != collapsed[-1]["rate"]:
            collapsed.append(tier)
    return collapsed



"""Build URDB-style output from Arcadia tariff data for a conversion scenario."""

import logging
from typing import cast

import questionary

from tariff_fetch.arcadia.api import ArcadiaSignalAPI
from tariff_fetch.urdb.arcadia.library import Library
from tariff_fetch.urdb.schema import URDBRate

from .energyschedule import build_energy_schedule
from .fixedcharge import build_fixed_charge
from .metadata import build_metadata
from .scenario import Scenario

logger = logging.getLogger(__name__)


def build_urdb(api: ArcadiaSignalAPI, scenario: Scenario) -> URDBRate:
    """Build a URDB record by combining energy, fixed-charge, and metadata chunks."""

    library = Library(api)
    try:
        energy_schedule = build_energy_schedule(scenario, library)
    except Exception as e:
        energy_schedule = _confirm_proceed(e, "energy rate strucutre")

    try:
        fixed_charge = build_fixed_charge(scenario, library)
    except Exception as e:
        fixed_charge = _confirm_proceed(e, "fixed charges")

    try:
        metadata = build_metadata(scenario, library)
    except Exception as e:
        metadata = _confirm_proceed(e, "metadata")

    if hasattr(library, "iter_issues"):
        for issue in library.iter_issues():
            logger.warning(issue)

    return {**energy_schedule, **fixed_charge, **metadata}


def _confirm_proceed(e: Exception, processing: str) -> URDBRate:
    """Ask whether to continue after a chunk-level conversion failure."""

    response = cast(
        bool | None,
        questionary.confirm(f"Error while converting: {processing}: {e}. Continue or print traceback and exit?").ask(),
    )
    if response is None:
        exit()
    if response:
        return {}
    raise e from None



"""Shared Arcadia tariff filtering and time applicability helpers."""

from collections.abc import Collection, Iterator
from datetime import date, datetime
from math import inf

from tariff_fetch.arcadia.schema.common import RateChargeClass, RateUnit
from tariff_fetch.arcadia.schema.season import SeasonExtended
from tariff_fetch.arcadia.schema.tariff import TariffExtended
from tariff_fetch.arcadia.schema.tariffrate import TariffRateBand, TariffRateExtended
from tariff_fetch.arcadia.schema.timeofuse import Period, TimeOfUseExtended
from tariff_fetch.urdb.arcadia.library import Library

from .exception import RateConversionError, TariffAccessDenied
from .scenario import Scenario
from .shared import as_naive_datetime

# ================================
# Tariff
# ================================


def tariff_iter_rates_for_dt(
    tariff: TariffExtended,
    scenario: Scenario,
    library: Library,
    dt: datetime,
) -> Iterator[TariffRateExtended]:
    """Yield all rate entries that apply for a scenario at a given datetime."""

    rates = tariff.get("rates", [])
    for rate in rates:
        if not rate_is_applied_to_scenario(rate, scenario, library):
            continue
        _record_calendar_issues(rate, library)
        if not rate_is_applied_to_datetime(rate, dt):
            continue
        if rate["rate_bands"]:
            yield rate
        elif rider_id := rate.get("rider_id"):
            try:
                rider_tariff = library.tariffs.get_tariff(rider_id)
            except TariffAccessDenied:
                library.record_issue(
                    ("inaccessible_rider", rate["tariff_rate_id"], rider_id),
                    f"Skipping inaccessible rider {rider_id} attached to rate {rate['tariff_rate_id']} ({rate['rate_name']})",
                )
                continue
            yield from tariff_iter_rates_for_dt(rider_tariff, scenario, library, dt)


# ================================
# Rate
# ================================


def rate_is_applied_to_scenario(
    rate: TariffRateExtended,
    scenario: Scenario,
    library: Library,
) -> bool:
    """Return whether a rate matches scenario-level filters such as territory and charge class."""

    if not rate_is_applied_to_charge_classes(rate, scenario.charge_classes):
        return False
    if (territory := rate.get("territory")) is not None:
        territory_id = territory["territory_id"]
        territory_ids = library.get_choice_property_as_ints("territoryId")
        if territory_id not in territory_ids:
            return False

    return True


def rate_is_applied_to_charge_classes(rate: TariffRateExtended, charge_classes: Collection[RateChargeClass]) -> bool:
    """Return whether a rate overlaps the requested Arcadia charge classes."""

    if (rate_charge_classes := rate.get("charge_class")) is None:
        return True
    return not set(rate_charge_classes).isdisjoint(charge_classes)


def rate_is_applied_to_datetime(rate: TariffRateExtended, dt: datetime) -> bool:
    """Return whether a rate is active at the given datetime."""

    dt_naive = as_naive_datetime(dt)
    if (from_dt := rate.get("from_date_time")) is not None and dt_naive < as_naive_datetime(from_dt):
        return False
    if (to_dt := rate.get("to_date_time")) is not None and dt_naive >= as_naive_datetime(to_dt):
        return False
    if (season := rate.get("season")) and not season_is_datetime_within(season, dt):
        return False
    if (tou := rate.get("time_of_use")) and not time_of_use_is_datetime_within(tou, dt):
        return False
    return True


def rate_filter_bands(rate: TariffRateExtended, _scenario: Scenario, library: Library) -> list[TariffRateBand]:
    """Return the supported rate bands whose applicability matches the current inputs."""

    if rate.get("variable_limit_key") is not None:
        raise RateConversionError(rate, "Rates with variable_limit_key are not supported")
    result: list[TariffRateBand] = []
    for band in rate.get("rate_bands"):
        if band.get("has_demand_limit"):
            raise RateConversionError(rate, "Bands with demand limits are not supported")
        if band.get("demand_upper_limit"):
            raise RateConversionError(rate, "Bands with demand limits are not supported")
        if band.get("has_property_limit"):
            raise RateConversionError(rate, "Bands with property limits are not supported")
        if band.get("property_upper_limit"):
            raise RateConversionError(rate, "Bands with property limits are not supported")
        if band.get("prev_upper_limit"):
            raise RateConversionError(rate, "Bands with property prev_upper_limit are not supported")
        if band.get("applicability_formula"):
            raise RateConversionError(rate, "Bands with property applicability_formula are not supported")
        if (applicability_value := band.get("applicability_value")) is not None:
            if (applicability_key := rate.get("applicability_key")) is None:
                raise RateConversionError(rate, "Band has applicability value but rate doesn't have applicability key")
            tariff_property = library.tariffs.get_property(applicability_key)

            if (tariff_property.get("period")) is not None:
                raise RateConversionError(rate, "Period properties are not supported for rate bands")

            if (category := tariff_property["property_types"]) not in {"APPLICABILITY", "RATE_CRITERIA"}:
                raise RateConversionError(rate, f"{category} is not a supported property category for tariff bands")

            operator = tariff_property["operator"]
            if operator != "=":
                raise RateConversionError(rate, "Only `=` operators are supported")
            match tariff_property["data_type"]:
                case "CHOICE":
                    if applicability_value not in library.get_property(applicability_key, "CHOICE"):
                        continue
                case "STRING":
                    if applicability_value != library.get_property(applicability_key, "STRING"):
                        continue
                case "BOOLEAN":
                    match applicability_value:
                        case "true":
                            applicability_value_bool = True
                        case "false":
                            applicability_value_bool = False
                        case _:
                            raise RateConversionError(
                                rate, "Boolean applicability value must be either `true` or `false`"
                            )
                    if applicability_value_bool != library.get_property(applicability_key, "BOOLEAN"):
                        continue
                case _:
                    raise RateConversionError(rate, "Unsupported data type for band applicabiltiy properties")
        result.append(band)
    return result


def rate_get_band_units(rate: TariffRateExtended) -> set[RateUnit]:
    """Return the set of rate-unit values used by a rate's bands."""

    return {b["rate_unit"] for b in rate.get("rate_bands", [])}


# ================================
# Rate Band
# ================================
def band_consumption_upper_limit(band: TariffRateBand) -> float:
    """Return a band's upper consumption limit, defaulting to infinity."""

    return band.get("consumption_upper_limit") or inf


def rate_band_get_amount_at_datetime(band: TariffRateBand, library: Library, dt: datetime) -> float:
    """Resolve the signed monetary amount for one rate band at a given datetime."""

    rate_id = band["tariff_rate_id"]
    rate = library.tariffs.get_rate(rate_id)
    if rate.get("variable_rate_sub_key") is not None:
        raise RateConversionError(rate, "Rates with variable_rate_sub_key are not supported")
    if rate.get("variable_factor_key") is not None:
        raise RateConversionError(rate, "Rates with variable_factor_key are not supported")
    mp = -1 if band["is_credit"] else 1
    if (variable_rate_key := rate.get("variable_rate_key")) is None:
        raw_result = band["rate_amount"] * mp
    else:
        raw_result = library.variables.lookup(variable_rate_key, dt) * mp

    calculation_factor = band.get("calculation_factor", 1.0)
    return raw_result * calculation_factor


# ================================
# Time of Use
# ================================


def time_of_use_is_datetime_within(
    tou: TimeOfUseExtended,
    dt: datetime,
) -> bool:
    """Return whether a datetime falls within a supported Arcadia TOU definition."""

    season = tou.get("season")
    if season and not season_is_datetime_within(season, dt):
        return False

    return any(period_is_datetime_within(period, dt) for period in tou["tou_periods"])


# ================================
# Time of Use - Period
# ================================


def period_is_datetime_within(
    period: Period,
    dt: datetime,
) -> bool:
    """Return whether a datetime falls within a TOU period's weekday and time bounds."""

    current_day = dt.weekday()
    start_day, end_day = period["from_day_of_week"], period["to_day_of_week"]

    if start_day <= end_day:
        if not (start_day <= current_day <= end_day):
            return False
    elif not (start_day <= current_day or current_day <= end_day):
        return False

    current_time = dt.hour * 60 + dt.minute
    start_time = period["from_hour"] * 60 + period["from_minute"]
    end_time = period["to_hour"] * 60 + period["to_minute"]

    if start_time == end_time:
        return True
    if start_time < end_time:
        return start_time <= current_time < end_time
    return current_time >= start_time or current_time < end_time


def _record_calendar_issues(rate: TariffRateExtended, library: Library) -> None:
    """Record ignored TOU calendar usage once per conversion run."""

    if (tou := rate.get("time_of_use")) is None:
        return
    if (calendar_id := tou.get("calendar_id")) is not None:
        library.record_issue(
            ("ignored_tou_calendar", rate["tariff_rate_id"], calendar_id),
            f"Ignoring TOU calendar_id {calendar_id} for rate {rate['tariff_rate_id']} ({rate['rate_name']})",
        )
    for period in tou.get("tou_periods", []):
        if (calendar_id := period.get("calendar_id")) is None:
            continue
        period_id = period.get("tou_period_id", "unknown")
        library.record_issue(
            ("ignored_tou_period_calendar", rate["tariff_rate_id"], period_id, calendar_id),
            f"Ignoring TOU period calendar_id {calendar_id} for rate {rate['tariff_rate_id']} ({rate['rate_name']})",
        )


# ================================
# Season
# ================================


def season_is_datetime_within(
    season: SeasonExtended,
    dt: datetime | date,
) -> bool:
    """Return whether a date or datetime falls within a seasonal date window."""

    dt = dt.date() if isinstance(dt, datetime) else dt
    start_month, end_month = season["season_from_month"], season["season_to_month"]
    start_day, end_day = season["season_from_day"], season["season_to_day"]
    start_date = date(dt.year, start_month, start_day)
    end_date = date(dt.year, end_month, end_day)
    if start_date < end_date:
        return start_date <= dt < end_date
    return start_date <= dt or dt < end_date



import json
import os
from pathlib import Path
from typing import cast

import questionary
import tenacity
from dotenv import load_dotenv
from fuzzywuzzy import fuzz  # pyright: ignore[reportMissingTypeStubs]
from selenium.common.exceptions import WebDriverException

from tariff_fetch._cli.types import Utility
from tariff_fetch.rateacuity import LoginState, create_context
from tariff_fetch.rateacuity.schema import Tariff

from . import console, prompt_filename


def process_rateacuity_gas(output_folder: Path, state: str):
    _ = load_dotenv()
    if not (username := os.getenv("RATEACUITY_USERNAME")):
        console.print("[b]RATEACUITY_USERNAME[/] environment variable is not set")
    if not (password := os.getenv("RATEACUITY_PASSWORD")):
        console.print("[b]RATEACUITY_PASSWORD[/] environment variable is not set")
    if not (username and password):
        console.print("Cannot use RateAcuity due to missing credentials")
        _ = console.input("Press enter to proceed...")
        return

    selected_utility = None
    tariffs_to_include = None
    results: list[Tariff] = []

    for attempt in tenacity.Retrying(
        stop=tenacity.stop_after_attempt(3), retry=tenacity.retry_if_exception_type(WebDriverException)
    ):
        with attempt, create_context() as context:
            with console.status("Fetching list of utilities..."):
                scraping_state = (
                    LoginState(context).login(username, password).gas().benchmark_all().select_state(state.upper())
                )
                utilities = [_ for _ in scraping_state.get_utilities() if _]

            if not utilities:
                raise RuntimeError(f"Something's wrong: rateacuity shows no utilities for this state ({state})")

            if selected_utility is None:
                selected_utility = cast(
                    str,
                    questionary.select(
                        message="Select a utility from available choices",
                        choices=utilities,
                        use_jk_keys=False,
                        use_search_filter=True,
                        use_shortcuts=False,
                    ).ask(),
                )
                if not selected_utility:
                    return

            with console.status("Fetching list of tariffs..."):
                scraping_state = scraping_state.select_utility(selected_utility)
                tariffs = [_ for _ in scraping_state.get_schedules() if _]

            if tariffs_to_include is None:
                tariffs_to_include = cast(
                    list[str],
                    questionary.checkbox(
                        message="Select tariffs to include",
                        choices=tariffs,
                        use_jk_keys=False,
                        use_search_filter=True,
                        validate=lambda _: bool(_) or "Select at least one tariff",
                    ).ask(),
                )

            if not tariffs_to_include:
                console.print("[red]No tariffs selected[/]")
                _ = console.input("Press enter to proceed...")
                return

            with console.status("Fetching tariffs..."):
                while tariffs_to_include:
                    tariff = tariffs_to_include.pop(0)
                    console.log(f"Fetching {tariff}")
                    scraping_state = scraping_state.select_schedule(tariff)
                    sections = scraping_state.as_sections()
                    results.append({"schedule": tariff, "sections": sections})
                    scraping_state = scraping_state.back_to_selections()

    assert selected_utility
    suggested_filename = f"gas_rateacuity_{selected_utility}"
    filename = prompt_filename(output_folder, suggested_filename, "json")
    filename.parent.mkdir(exist_ok=True)
    _ = filename.write_text(json.dumps(results, indent=2))


def process_rateacuity(output_folder: Path, state: str, utility: Utility):
    _ = load_dotenv()
    if not (username := os.getenv("RATEACUITY_USERNAME")):
        console.print("[b]RATEACUITY_USERNAME[/] environment variable is not set")
    if not (password := os.getenv("RATEACUITY_PASSWORD")):
        console.print("[b]RATEACUITY_PASSWORD[/] environment variable is not set")
    if not (username and password):
        console.print("Cannot use RateAcuity due to missing credentials")
        _ = console.input("Press enter to proceed...")
        return

    selected_utility = None
    tariffs_to_include = None
    results: list[Tariff] = []

    for attempt in tenacity.Retrying(
        stop=tenacity.stop_after_attempt(3), retry=tenacity.retry_if_exception_type(WebDriverException)
    ):
        with attempt, create_context() as context:
            with console.status("Fetching list of utilities..."):
                scraping_state = (
                    LoginState(context).login(username, password).electric().benchmark_all().select_state(state.upper())
                )
                utilities = [_ for _ in scraping_state.get_utilities() if _]

            if not utilities:
                raise RuntimeError(f"Something's wrong: rateacuity shows no utilities for this state ({state})")

            if selected_utility is None:
                utilities_scored = sorted(utilities, key=lambda _: fuzz.ratio(utility.name, _), reverse=True)  # pyright: ignore[reportUnknownMemberType]
                selected_utility = utilities_scored.pop(0)
                if not questionary.confirm(f"Is this the correct utility: {selected_utility} ?").ask():
                    selected_utility = cast(
                        str,
                        questionary.select(
                            message="Select a utility from available choices",
                            choices=utilities_scored,
                            use_jk_keys=False,
                            use_search_filter=True,
                            use_shortcuts=False,
                        ).ask(),
                    )
                if not selected_utility:
                    return

            with console.status("Fetching list of tariffs..."):
                scraping_state = scraping_state.select_utility(selected_utility)
                tariffs = [_ for _ in scraping_state.get_schedules() if _]

            if tariffs_to_include is None:
                tariffs_to_include = cast(
                    list[str],
                    questionary.checkbox(
                        message="Select tariffs to include",
                        choices=tariffs,
                        use_jk_keys=False,
                        use_search_filter=True,
                        validate=lambda _: bool(_) or "Select at least one tariff",
                    ).ask(),
                )

            if not tariffs_to_include:
                console.print("[red]No tariffs selected[/]")
                _ = console.input("Press enter to proceed...")
                return

            with console.status("Fetching tariffs..."):
                while tariffs_to_include:
                    tariff = tariffs_to_include.pop(0)
                    console.log(f"Fetching {tariff}")
                    scraping_state = scraping_state.select_schedule(tariff)
                    sections = scraping_state.as_sections()
                    results.append({"schedule": tariff, "sections": sections})
                    scraping_state = scraping_state.back_to_selections()

    assert selected_utility
    suggested_filename = f"rateacuity_{selected_utility}"
    if not (filename := prompt_filename(output_folder, suggested_filename, "json")):
        return
    filename.parent.mkdir(exist_ok=True)
    _ = filename.write_text(json.dumps(results, indent=2))



from typing import Annotated, NotRequired

from pydantic import Field
from pydantic.alias_generators import to_camel
from typing_extensions import TypedDict

from .common import SeasonPredominance


class SeasonStandardFields(TypedDict):
    season_id: int
    """Unique Arcadia ID (primary key) for each season"""
    lse_id: int
    """The ID of the LSE this territory belongs to"""
    season_group_id: int
    """The ID of the season group that contains this season"""
    season_name: str
    """The name of the season (i.e. "Summer" or "Winter")"""
    season_from_month: Annotated[int, Field(ge=1, le=12)]
    """	Value of 1-12 representing the month this season begins"""
    season_from_day: Annotated[int, Field(ge=1, le=31)]
    """Value of 1-31 (depending on month) representing the day this season begins"""
    season_to_month: Annotated[int, Field(ge=1, le=12)]
    """Value of 1-12 representing the month this season ends"""
    season_to_day: Annotated[int, Field(ge=1, le=31)]
    """Value of 1-31 (depending on month) representing the day this season ends"""


class SeasonExtendedFields(TypedDict):
    from_edge_predominance: NotRequired[SeasonPredominance | None]
    """
    Can be None (from date is a fixed date),
    PREDOMINANT (from date is the start of the bill the season starts in),
    or SUBSERVIENT (from date is the end of the bill the season starts in).
    """
    to_edge_predominance: NotRequired[SeasonPredominance | None]
    """
	Can be None (to date is a fixed date),
    PREDOMINANT (to date is the start of the bill the season starts ends in),
    or SUBSERVIENT (to date is the end of the bill the season starts in).
    """


class SeasonStandard(SeasonStandardFields):
    __pydantic_config__ = {"alias_generator": to_camel, "extra": "forbid"}  # pyright: ignore[reportGeneralTypeIssues, reportUnannotatedClassAttribute]  # noqa


class SeasonExtended(SeasonStandardFields, SeasonExtendedFields):
    __pydantic_config__ = {"alias_generator": to_camel, "extra": "forbid"}  # pyright: ignore[reportGeneralTypeIssues, reportUnannotatedClassAttribute]  # noqa



"""Scenario inputs that control one Arcadia-to-URDB conversion run."""

from dataclasses import dataclass, field
from typing import cast, get_args

from tariff_fetch.arcadia.schema.common import RateChargeClass

_CHARGE_CLASSES = cast(tuple[RateChargeClass, ...], get_args(RateChargeClass))


@dataclass(frozen=True)
class Scenario:
    """User-selected inputs for converting one Arcadia master tariff."""

    master_tariff_id: int
    year: int
    apply_percentages: bool
    charge_classes: set[RateChargeClass] = field(default_factory=set)



from datetime import datetime
from math import inf
from types import SimpleNamespace

from tariff_fetch.urdb.arcadia import energyschedule as es
from tariff_fetch.urdb.arcadia import fixedcharge as fc
from tariff_fetch.urdb.arcadia import metadata as metadata_mod
from tariff_fetch.urdb.arcadia.scenario import Scenario
from tests.arcadia_urdb_fixtures import make_band, make_fixed_rate, make_library_with_tariff


def test_golden_flat_energy_schedule(monkeypatch):
    monkeypatch.setattr(es, "get_month_hour_bands", lambda scenario, library, month, hour, day_filter: [(inf, 0.15)])

    result = es.build_energy_schedule(Scenario(1, 2025, apply_percentages=False), library=None)  # type: ignore[arg-type]

    assert result == {
        "energyratestructure": [[{"rate": 0.15, "unit": "kWh"}]],
        "energyweekdayschedule": tuple(tuple(0 for _ in range(24)) for _ in range(12)),
        "energyweekendschedule": tuple(tuple(0 for _ in range(24)) for _ in range(12)),
    }


def test_golden_tiered_energy_schedule(monkeypatch):
    monkeypatch.setattr(
        es,
        "get_month_hour_bands",
        lambda scenario, library, month, hour, day_filter: [(100.0, 0.1), (inf, 0.2)],
    )

    result = es.build_energy_schedule(Scenario(1, 2025, apply_percentages=False), library=None)  # type: ignore[arg-type]

    assert result == {
        "energyratestructure": [[{"rate": 0.1, "max": 100.0, "unit": "kWh"}, {"rate": 0.2, "unit": "kWh"}]],
        "energyweekdayschedule": tuple(tuple(0 for _ in range(24)) for _ in range(12)),
        "energyweekendschedule": tuple(tuple(0 for _ in range(24)) for _ in range(12)),
    }


def test_golden_fixed_charge_build(monkeypatch):
    fixed_rate = make_fixed_rate(rate_bands=[make_band(rate_amount=12.5)])
    library = make_library_with_tariff(fixed_rate)

    monkeypatch.setattr(fc.ru, "tariff_iter_rates_for_dt", lambda tariff, scenario, library, dt: [fixed_rate])
    monkeypatch.setattr(fc.ru, "rate_filter_bands", lambda rate, scenario, library: list(rate["rate_bands"]))
    monkeypatch.setattr(fc.ru, "rate_band_get_amount_at_datetime", lambda band, library, dt: band["rate_amount"])

    result = fc.build_fixed_charge(Scenario(1, 2025, apply_percentages=False), library)  # type: ignore[arg-type]

    assert result == {
        "fixedchargefirstmeter": 12.5,
        "fixedchargeunits": "$/month",
    }


def test_golden_daily_fixed_charge_build(monkeypatch):
    fixed_rate = make_fixed_rate(charge_period="DAILY", rate_bands=[make_band(rate_amount=1.0)])
    library = make_library_with_tariff(fixed_rate)

    monkeypatch.setattr(fc.ru, "tariff_iter_rates_for_dt", lambda tariff, scenario, library, dt: [fixed_rate])
    monkeypatch.setattr(fc.ru, "rate_filter_bands", lambda rate, scenario, library: list(rate["rate_bands"]))
    monkeypatch.setattr(fc.ru, "rate_band_get_amount_at_datetime", lambda band, library, dt: band["rate_amount"])

    result = fc.get_fixed_charge_at_dt(
        Scenario(1, 2025, apply_percentages=False),
        library,  # type: ignore[arg-type]
        datetime(2025, 4, 1, 0, 30),
    )

    assert result == 30.0


def test_golden_metadata_chunk():
    library = SimpleNamespace(
        tariffs=SimpleNamespace(
            get_tariff_at_date=lambda master_tariff_id, dt: {
                "lse_code": "CONED",
                "lse_name": "Consolidated Edison",
                "tariff_name": "Service Classification 1",
            }
        )
    )

    result = metadata_mod.build_metadata(Scenario(2252, 2025, apply_percentages=False), library)  # type: ignore[arg-type]

    assert result == {
        "label": "CONED",
        "utility": "Consolidated Edison",
        "name": "Service Classification 1",
        "country": "USA",
    }



from typing import NotRequired

from pydantic.alias_generators import to_camel
from typing_extensions import TypedDict

from .common import TerritoryItemType, TerritoryUsageType


class Coordinate(TypedDict):
    latitude: float
    longitude: float


class TerritoryItem(TypedDict):
    territory_item_id: int
    """	Unique Arcadia ID (primary key) for each territory Item"""
    territory_type: TerritoryItemType
    value: str
    """The name of the territory item (i.e. 'Kirkwood' or '94115')."""
    exclude: bool
    """If True, this territory item is not included as part of the coverage area of the parent territory."""
    partial: bool
    """If True, then only a partial area of this territory item is covered by this LSE."""


class TerritoryLSE(TypedDict):
    territory_id: int
    """	The territoryId this Territory LSE belongs to"""
    lse_id: int
    """The lseId of the LSE offering service in this Territory"""
    lse_name: str
    """The name of the LSE"""
    distribution: bool
    """indicates whether this LSE is a distribution LSE"""
    supplier_residential: bool
    """indicates whether this LSE supplies service to residential customers"""
    supplier_general: bool
    """indicates whether this LSE supplies service to general (commercial and industrial) customers"""
    residential_coverage: float
    """Percentage of residential customers this LSE supplies service to"""
    general_coverage: float
    """Percentage of general (commercial and industrial) customers this LSE supplies service to"""


class TerritoryMinimalFields(TypedDict):
    territory_id: int
    """Unique Arcadia ID (primary key) for each territory"""
    territory_name: str
    """The name of the territory (i.e. 'Service Area for CA' or 'Baseline Region H')."""
    lse_id: int
    """The ID of the LSE this Territory belongs to"""
    lse_name: str
    """The name of the LSE (utility) that this territory belongs to."""
    country_code: str


class TerritoryStandardFields(TypedDict):
    parent_territory_id: int | None
    """The ID of the parent territory

    Typically this will be on a tariff territory that ties it back to the service territory.
    """
    usage_type: TerritoryUsageType
    """The type of the items that define the physical area of coverage of this territory"""
    item_types: TerritoryItemType
    items: NotRequired[list[TerritoryItem]]
    """The list of Territory Items that define the area covered by this territory."""
    territory_lses: NotRequired[list[TerritoryLSE]]
    """	The list of LSEs that offer service (retail) to customers in this territory. Applies to deregulated markets."""
    dereg_res: bool
    """Whether the residential electricity market in this territory is deregulated."""
    dereg_candi: bool
    """	Whether the commercial and industrial electricity market in this territory is deregulated."""
    center_point: Coordinate
    """	The latitude and longitude of the centerPoint of this territory."""


class TerritoryMinimal(TerritoryMinimalFields):
    __pydantic_config__ = {"alias_generator": to_camel, "extra": "forbid"}  # pyright: ignore[reportGeneralTypeIssues, reportUnannotatedClassAttribute]  # noqa


class TerritoryStandard(TerritoryMinimalFields, TerritoryStandardFields):
    __pydantic_config__ = {"alias_generator": to_camel, "extra": "forbid"}  # pyright: ignore[reportGeneralTypeIssues, reportUnannotatedClassAttribute]  # noqa



import os
from collections.abc import Iterator
from dataclasses import dataclass, field
from datetime import date
from json import JSONDecodeError
from typing import Annotated, Any, Generic, Literal, Self, TypeVar, Unpack, overload
from urllib.parse import urljoin

import requests
from pydantic import ConfigDict, PlainSerializer, TypeAdapter
from pydantic.alias_generators import to_camel
from requests.auth import HTTPBasicAuth
from typing_extensions import TypedDict

from .schema.common import CustomerClass, TariffType
from .schema.lookup import Lookup
from .schema.lse import LSEExtended, LSEMinimal, ServiceType
from .schema.tariff import TariffExtended, TariffStandard

_BASE_URL = "https://api.genability.com/rest/public/"
_comma_separated = PlainSerializer(",".join)

_T = TypeVar("_T")
_C = TypeVar("_C")


class SearchParams(TypedDict, total=False):
    search: str
    """
    The string of text to search on.
    This can also be a regular expression, in which case you should
    set the isRegex flag to true (see below).
    """
    search_on: Annotated[list[str], _comma_separated]
    """
    Comma-separated list of fields to query on. When searchOn is specified, the text provided
    in the search string field will be searched within these fields. The list of fields to search on
    depends on the entity being searched for. Read the documentation for the entity for more details
    """
    starts_with: bool
    "When true, the search will only return results that begin with the specified search string."
    ends_with: bool
    "When true, the search will only return results that end with the specified search string."
    is_regex: bool
    """
    When true, the provided search string will be regarded as a regular expression
    and the search will return results matching the regular expression.
    Default is False.
    """
    sort_on: Annotated[list[str], _comma_separated]
    sort_order: Annotated[list[Literal["ASC", "DESC"]], _comma_separated]
    """
    Comma-separated list of ordering. Possible values are ASC and DESC. Default is ASC.
    If your sortOn contains multiple fields and you would like to order fields individually,
    you can pass in a comma-separated list here
    """


class PagingParams(TypedDict, total=False):
    __pydantic_config__ = ConfigDict(alias_generator=to_camel)  # pyright: ignore[reportGeneralTypeIssues,reportUnannotatedClassAttribute]
    page_start: Annotated[int, PlainSerializer(str)]
    page_count: Annotated[int, PlainSerializer(str)]


class GetLSEsParams(TypedDict, total=False):
    __pydantic_config__ = ConfigDict(alias_generator=to_camel)  # pyright: ignore[reportGeneralTypeIssues,reportUnannotatedClassAttribute]
    zip_code: str
    "Zip or post code where you would like to see a list of LSEs for (e.g. 5 digit ZIP code for USA)."
    country: str
    "ISO country code"
    ownerships: Annotated[list[Literal["INVESTOR", "COOP", "MUNI"]], _comma_separated]
    "Filter results by the type of ownership structure for the LSE."
    serviceTypes: Annotated[list[ServiceType], _comma_separated]
    "Filter results to LSEs that just offer this service type to a customer class"
    residential_service_types: Annotated[list[ServiceType], _comma_separated]
    commercial_service_types: Annotated[list[ServiceType], _comma_separated]
    industrial_service_types: Annotated[list[ServiceType], _comma_separated]
    transportation_service_types: Annotated[list[ServiceType], _comma_separated]


class TariffsParams(TypedDict, total=False):
    lse_id: int
    effective_on: date
    customer_classes: Annotated[list[CustomerClass], _comma_separated]
    master_tariff_id: int
    tariff_types: Annotated[list[TariffType], _comma_separated]
    populate_properties: bool
    populate_rates: bool
    populate_documents: bool
    from_date_time: date
    to_date_time: date


class LookupsParams(PagingParams, total=False):
    __pydantic_config__ = ConfigDict(alias_generator=to_camel)  # pyright: ignore[reportGeneralTypeIssues,reportUnannotatedClassAttribute]
    sub_property_key: str
    """Subproperty key name"""
    from_date_time: date
    to_date_time: date


class PropertyParams(TypedDict, total=False):
    __pydantic_config__ = ConfigDict(alias_generator=to_camel)  # pyright: ignore[reportGeneralTypeIssues,reportUnannotatedClassAttribute]


class PageResponse(TypedDict, Generic[_T]):
    __pydantic_config__ = {"alias_generator": to_camel, "extra": "forbid"}  # pyright: ignore[reportGeneralTypeIssues, reportUnannotatedClassAttribute]  # noqa
    status: str
    count: int
    type: str
    results: list[_T]
    page_count: int
    page_start: int


class GetLSEsPageParams(PagingParams, SearchParams, GetLSEsParams):
    __pydantic_config__ = {"alias_generator": to_camel, "extra": "forbid"}  # pyright: ignore[reportGeneralTypeIssues, reportUnannotatedClassAttribute]  # noqa


class GetTariffsPageParams(PagingParams, SearchParams, TariffsParams):
    __pydantic_config__ = {"alias_generator": to_camel, "extra": "forbid"}  # pyright: ignore[reportGeneralTypeIssues, reportUnannotatedClassAttribute]  # noqa


@dataclass(frozen=True)
class ArcadiaSignalAPIAuth:
    app_id: str
    app_key: str

    @classmethod
    def from_env(cls) -> Self:
        app_id = os.getenv("ARCADIA_APP_ID")
        if app_id is None:
            raise ValueError("ARCADIA_APP_ID variable is not set. Either set it or use auth as a parameter.")
        app_key = os.getenv("ARCADIA_APP_KEY")
        if app_key is None:
            raise ValueError("ARCADIA_APP_KEY variable is not set. Either set it or use auth as a parameter.")
        return cls(app_id, app_key)


@dataclass(frozen=True)
class ArcadiaSignalAPI:
    auth: ArcadiaSignalAPIAuth = field(default_factory=ArcadiaSignalAPIAuth.from_env)
    base_url: str = _BASE_URL
    session: requests.Session = field(default_factory=requests.Session)

    def _request(self, path: str, **params) -> dict[str, Any]:  # pyright: ignore[reportMissingParameterType, reportUnknownParameterType, reportExplicitAny]
        response = self.session.get(
            urljoin(self.base_url, path),
            params=params,  # pyright: ignore[reportUnknownArgumentType]
            auth=HTTPBasicAuth(self.auth.app_id, self.auth.app_key),
        )
        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            if (message := _extract_arcadia_error_message(response)) is not None:
                raise requests.HTTPError(message, response=response, request=response.request) from e
            raise
        return response.json()  # pyright: ignore[reportAny]

    def _request_typed(
        self,
        path: str,
        params: _C,
        extras: dict[str, Any],  # pyright: ignore[reportExplicitAny]
        params_type: type[_C],
        response_type: type[_T],
    ) -> _T:
        param_type_adapter = TypeAdapter(params_type)
        params_to_api = param_type_adapter.dump_python(params, by_alias=True)  # pyright: ignore[reportAny]
        response = self._request(path, **params_to_api, **extras)  # pyright: ignore[reportUnknownMemberType, reportAny]
        result_adapter = TypeAdapter(response_type)
        return result_adapter.validate_python(response)

    @property
    def lses(self) -> "ArcadiaSignalAPILse":
        return ArcadiaSignalAPILse(self.auth, self.base_url, self.session)

    @property
    def tariffs(self) -> "ArcadiaSignalAPITariff":
        return ArcadiaSignalAPITariff(self.auth, self.base_url, self.session)

    @property
    def properties(self) -> "ArcadiaPropertiesAPI":
        return ArcadiaPropertiesAPI(self.auth, self.base_url, self.session)


@dataclass(frozen=True)
class ArcadiaSignalAPILse(ArcadiaSignalAPI):
    @overload
    def get_page(self, **params: Unpack[GetLSEsPageParams]) -> PageResponse[LSEMinimal]: ...

    @overload
    def get_page(self, fields: None, **params: Unpack[GetLSEsPageParams]) -> PageResponse[LSEMinimal]: ...

    @overload
    def get_page(self, fields: Literal["min"], **params: Unpack[GetLSEsPageParams]) -> PageResponse[LSEMinimal]: ...

    @overload
    def get_page(self, fields: Literal["ext"], **params: Unpack[GetLSEsPageParams]) -> PageResponse[LSEExtended]: ...

    def get_page(self, fields: Literal["min", "ext"] | None = None, **params: Unpack[GetLSEsPageParams]):
        match fields:
            case "min" | None:
                return self._request_typed(
                    "lses", params, {"fields": "min"}, GetLSEsPageParams, PageResponse[LSEMinimal]
                )
            case "ext":
                return self._request_typed(
                    "lses", params, {"fields": "ext"}, GetLSEsPageParams, PageResponse[LSEExtended]
                )


@dataclass(frozen=True)
class ArcadiaSignalAPITariff(ArcadiaSignalAPI):
    # @overload
    # def get_page(
    #    self, fields: Literal["min"], **params: Unpack[GetTariffsPageParams]
    # ) -> PageResponse[TariffMinimal]: ...

    @overload
    def get_page(
        self, fields: Literal["ext"], **params: Unpack[GetTariffsPageParams]
    ) -> PageResponse[TariffExtended]: ...

    @overload
    def get_page(self, fields: None, **params: Unpack[GetTariffsPageParams]) -> PageResponse[TariffStandard]: ...

    @overload
    def get_page(self, **params: Unpack[GetTariffsPageParams]) -> PageResponse[TariffStandard]: ...

    def get_page(self, fields: Literal["ext"] | None = None, **params: Unpack[GetTariffsPageParams]):
        match fields:
            # case "min":
            # return self._request_typed(
            #    "tariffs", params, {"fields": "min"}, GetTariffsPageParams, PageResponse[TariffMinimal]
            # )
            case "ext":
                return self._request_typed(
                    "tariffs", params, {"fields": "ext"}, GetTariffsPageParams, PageResponse[TariffExtended]
                )
            case None:
                return self._request_typed("tariffs", params, {}, GetTariffsPageParams, PageResponse[TariffStandard])

    # @overload
    # def iter_pages(self, fields: Literal["min"], **params: Unpack[GetTariffsPageParams]) -> Iterator[TariffMinimal]: ...

    @overload
    def iter_pages(
        self, fields: Literal["ext"], **params: Unpack[GetTariffsPageParams]
    ) -> Iterator[TariffExtended]: ...

    @overload
    def iter_pages(self, fields: None, **params: Unpack[GetTariffsPageParams]) -> Iterator[TariffStandard]: ...

    @overload
    def iter_pages(self, **params: Unpack[GetTariffsPageParams]) -> Iterator[TariffStandard]: ...

    def iter_pages(
        self, fields: Literal["ext"] | None = None, **params: Unpack[GetTariffsPageParams]
    ) -> Iterator[TariffExtended] | Iterator[TariffStandard]:
        page_start = params.get("page_start", 0)
        while True:
            params_: GetTariffsPageParams = {**params, "page_start": page_start}
            response = self.get_page(fields, **params_)
            count, page_start = response["count"], response["page_start"]
            size = len(response["results"])
            yield from response["results"]  # pyright: ignore[reportReturnType]
            if page_start + size >= count:
                return
            if size <= 0:
                return
            page_start += size


@dataclass(frozen=True)
class ArcadiaPropertiesAPI(ArcadiaSignalAPI):
    @property
    def lookups(self) -> "ArcadiaLookupsAPI":
        return ArcadiaLookupsAPI(self.auth, self.base_url, self.session)


@dataclass(frozen=True)
class ArcadiaLookupsAPI(ArcadiaPropertiesAPI):
    def get_page(self, property_key: str, **params: Unpack[LookupsParams]) -> PageResponse[Lookup]:
        return self._request_typed(
            f"properties/{property_key}/lookups", params, {}, LookupsParams, PageResponse[Lookup]
        )

    def iter_pages(self, property_key: str, **params: Unpack[LookupsParams]) -> Iterator[Lookup]:
        page_start = params.get("page_start", 0)
        while True:
            params_: LookupsParams = {**params, "page_start": page_start}
            response = self.get_page(property_key, **params_)
            count, page_start = response["count"], response["page_start"]
            size = len(response["results"])
            yield from response["results"]
            if page_start + size >= count:
                return
            if size <= 0:
                return
            page_start += size


def _extract_arcadia_error_message(response: requests.Response) -> str | None:
    try:
        payload = response.json()  # pyright: ignore[reportAny]
    except (JSONDecodeError, ValueError):
        return None

    if not isinstance(payload, dict):
        return None

    results = payload.get("results")  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]
    if not isinstance(results, list):
        return None

    messages = [  # pyright: ignore[reportUnknownVariableType]
        item["message"]
        for item in results  # pyright: ignore[reportUnknownVariableType]
        if isinstance(item, dict) and isinstance(item.get("message"), str) and item["message"].strip()  # pyright: ignore[reportUnknownMemberType]
    ]
    if not messages:
        return None

    return "; ".join(messages)  # pyright: ignore[reportUnknownArgumentType]



import argparse
import datetime
import json
import os
from collections.abc import Iterator, Sequence
from pathlib import Path
from typing import Any, Literal, TypeAlias, cast

from typing_extensions import TypedDict, Unpack

from .base import api_request_json  # pyright: ignore[reportUnknownVariableType]

__all__ = [
    "UTILITY_RATES_API_PATH",
    "AttributeList",
    "CoincidentRateStructure",
    "CoincidentRateTier",
    "DemandRateStructure",
    "DemandRateTier",
    "DemandUnit",
    "DistributedGenerationRule",
    "EnergyRateStructure",
    "EnergyRateTier",
    "FixedChargeUnit",
    "FlatDemandStructure",
    "FlatDemandTier",
    "PhaseWiring",
    "ScheduleMatrix",
    "ServiceType",
    "UtilityRateSector",
    "UtilityRatesParams",
    "UtilityRatesResponse",
    "UtilityRatesResponseItem",
    "VoltageCategory",
    "iter_utility_rates",
    "main",
    "utility_rates",
]


UTILITY_RATES_API_PATH = "/utility_rates"
DemandUnit: TypeAlias = (
    Literal["kW"] | Literal["hp"] | Literal["kVA"] | Literal["kW daily"] | Literal["hp daily"] | Literal["kVA daily"]
)
VoltageCategory: TypeAlias = Literal["Primary"] | Literal["Secondary"] | Literal["Transmission"]
PhaseWiring: TypeAlias = Literal["Single Phase"] | Literal["3-Phase"] | Literal["Single and 3-Phase"]
ServiceType: TypeAlias = (
    Literal["Bundled"] | Literal["Energy"] | Literal["Delivery"] | Literal["Delivery with Standard Offer"]
)
FixedChargeUnit: TypeAlias = Literal["$/day"] | Literal["$/month"] | Literal["$/year"]
DistributedGenerationRule: TypeAlias = Literal[
    "Net Metering", "Net Billing Instantaneous", "Net Billing Hourly", "Buy All Sell All"
]
UtilityRateSector: TypeAlias = (
    Literal["Residential"] | Literal["Commercial"] | Literal["Industrial"] | Literal["Lighting"]
)
ScheduleMatrix: TypeAlias = list[list[int]]
AttributeList: TypeAlias = list[dict[str, Any]]  # pyright: ignore[reportExplicitAny]


class FlatDemandTier(TypedDict, total=False):
    max: float
    rate: float
    adj: float


class DemandRateTier(TypedDict, total=False):
    max: float
    rate: float
    adj: float


class CoincidentRateTier(TypedDict, total=False):
    max: float
    rate: float
    adj: float


class EnergyRateTier(TypedDict, total=False):
    max: float
    unit: str
    rate: float
    adj: float
    sell: float


FlatDemandStructure: TypeAlias = list[list[FlatDemandTier]]
DemandRateStructure: TypeAlias = list[list[DemandRateTier]]
CoincidentRateStructure: TypeAlias = list[list[CoincidentRateTier]]
EnergyRateStructure: TypeAlias = list[list[EnergyRateTier]]


class UtilityRatesParams(TypedDict, total=False):
    """Parameters for utility rates api"""

    modified_after: datetime.datetime | int
    "Modified after, datetime or seconds since 1970-01-01T00:00:00, UTC. "
    limit: int
    "Maximum record count 500"
    getpage: str
    """
    Get a specific page name, i.e. getpage=535aeca39bef511109580ee1.
    This is referred to as 'label' in the results.
    """
    ratesforutility: str
    """
    Get rates for a specific utility page name, i.e. ratesforutility=Detroit Edison Co.
    This value can be re-used from the "label" value in the utility_companies results.
    """
    offset: int
    "The offset from which to start retrieving records."
    orderby: str
    "Field to use to sort results."
    direction: Literal["asc"] | Literal["desc"]
    "Direction to order results, ascending or descending."
    effective_on_date: datetime.datetime | int
    "Effective on date, datetime or seconds since 1970-01-01T00:00:00, UTC."
    sector: UtilityRateSector
    "Shows only those rates matching the given sector."
    approved: bool
    "Shows only those rates whose approval status matches the given value."
    is_default: bool
    "Shows only those rates whose 'Is Default' status matches the given value."
    country: str
    "ISO 3 character country code, i.e. country=USA."
    address: str
    "Address for rate, see Google Geocoding API for details. Address or lat/lon may be used, but not both."
    lat: int
    "Latitude for rate. If set, lon must also be set and address must not."
    lon: int
    "Longitude for rate. If set, lat must also be set and address must not."
    radius: int
    "Radius to include surrounding search result, in miles. (min 0, max 200)"
    co_limit: int
    "Maximum number of companies to include in a geographic search (using lat/lon, or address)."
    eia: int
    "EIA Id to look up."
    callback: str
    "callback=<mycallback> - set mycallback as the json callback."
    detail: Literal["full"] | Literal["minimal"]
    """
    detail=full - returns every variable. Since this results in a lot of data that can time-out
    returning to your server, use a limit=500 and set an offset (e.g. 501) if you want more data.
    """


class UtilityRatesResponseItem(TypedDict):
    label: str
    "Page label"
    utility: str
    "Utility company name"
    name: str
    "Rate name"
    uri: str
    "Full page URI"
    approved: bool
    "Expert has verified"
    is_default: bool
    "Most common rate for given time period, sector, and service type"
    startdate: int
    "Effective date timestamp"
    enddate: int
    "End date timestamp"
    supercedes: str
    "Label of the rate this rate supercedes."
    sector: UtilityRateSector
    "Sector"
    servicetype: ServiceType
    "Service type"
    description: str
    "Description"
    source: str
    "Source or reference."
    sourceparent: str
    "Source parent URI"
    basicinformationcomments: str
    "Basic comments"
    peakkwcapacitymin: float
    "Demand minimum (kW)"
    peakkwcapacitymax: float
    "Demand maximum (kW)"
    demandunits: DemandUnit
    "Demand applicability units"
    peakkwcapacityhistory: float
    "Demand history (months)"
    peakkwhusagemin: float
    "Energy minimum (kW)"
    peakkwhusagemax: float
    "Energy maximum (kW)"
    peakkwhusagehistory: float
    "Energy history (months)"
    voltageminimum: float
    "Service voltage minimum"
    voltagemaximum: float
    "Service voltage maximum"
    voltagecategory: VoltageCategory
    "Voltage category"
    phasewiring: PhaseWiring
    "Phase wiring"
    flatdemandunit: DemandUnit
    "Seasonal/monthly demand rate units"
    flatdemandstructure: FlatDemandStructure
    "Seasonal/monthly demand charge structure"
    flatdemandmonths: list[int]
    "Seasonal/monthly demand charge schedule"
    demandrateunit: DemandUnit
    "Time-of-use demand rate units"
    demandratestructure: DemandRateStructure
    "Time-of-use demand charge structure"
    demandweekdayschedule: ScheduleMatrix
    "Time-of-use demand weekday schedule"
    demandweekendschedule: ScheduleMatrix
    "Time-of-use demand weekend schedule"
    demandratchetpercentage: list[float]
    "Demand ratchet percentage by month"
    demandwindow: float
    "Demand window (minutes)"
    demandreactivepowercharge: float
    "Demand reactive power charge ($/kVAR)"
    coincidentrateunit: DemandUnit
    "Coincident demand rate units"
    coincidentratestructure: CoincidentRateStructure
    "Coincident demand charge structure"
    coincidentrateschedule: ScheduleMatrix
    "Coincident demand schedule"
    demandattrs: AttributeList
    "Demand attribute key/value pairs"
    demandcomments: str
    "Demand comments"
    dgrules: DistributedGenerationRule
    "Distributed generation compensation rules"
    energyratestructure: EnergyRateStructure
    "Tiered energy usage charge structure"
    energyweekdayschedule: ScheduleMatrix
    "Tiered energy weekday schedule"
    energyweekendschedule: ScheduleMatrix
    "Tiered energy weekend schedule"
    fueladjustmentsmonthly: list[float]
    "Fuel adjustments by month ($/kWh)"
    energyattrs: AttributeList
    "Energy attribute key/value pairs"
    energycomments: str
    "Energy comments"
    fixedchargefirstmeter: float
    "Fixed charge for first meter"
    fixedchargeeaaddl: float
    "Fixed charge for each additional meter"
    fixedchargeunits: FixedChargeUnit
    "Units for fixed charge values"
    mincharge: float
    "Minimum charge"
    minchargeunits: FixedChargeUnit
    "Units for minimum charge"
    fixedattrs: AttributeList
    "Fixed charge attribute key/value pairs"


class UtilityRatesResponse(TypedDict):
    items: list[UtilityRatesResponseItem]


def iter_utility_rates(
    api_key: str,
    iter_offset: int = 0,
    format: Literal["json"] | Literal["json_plain"] | Literal["csv"] = "json",
    version: str = "latest",
    **kwargs: Unpack[UtilityRatesParams],
) -> Iterator[UtilityRatesResponseItem]:
    offset = iter_offset
    rows = 1
    while rows:
        response = utility_rates(api_key, format, version, **{**kwargs, "offset": offset})
        items = response["items"]
        rows = len(items)
        offset += rows
        yield from items


def utility_rates(
    api_key: str,
    format: Literal["json"] | Literal["json_plain"] | Literal["csv"] = "json",
    version: str = "latest",
    **kwargs: Unpack[UtilityRatesParams],
) -> UtilityRatesResponse:
    """Access utility rate structure information from the U.S. Utility Rate Database

    Args:
        api_key: api key (register at https://openei.org/services/api/signup/)

        format: The format parameter will be disregarded if the debug parameter is set

        version: API version to use

          OpenEI retired the version 1 API on January 2nd, 2014.
          Version 2 was released on May 4th, 2013, and version 3 was released on June 16th, 2014,
          at which point version 2 database updates ceased, with all updates now appearing only
          in version 3 or greater.
    """
    return cast(  # pyright: ignore[reportInvalidCast]
        UtilityRatesResponse,
        api_request_json(path=UTILITY_RATES_API_PATH, api_key=api_key, format=format, version=version, **kwargs),
    )


def main(argv: Sequence[str] | None = None) -> None:
    parser = argparse.ArgumentParser(description="Fetch OpenEI utility rates and write them to a JSON file.")
    _ = parser.add_argument(
        "ratesforutility",
        help="Utility label (see OpenEI utility companies) to request rates for.",
    )
    _ = parser.add_argument(
        "--api-key",
        default=os.getenv("OPENEI_API_KEY"),
        help="OpenEI API key (defaults to OPENEI_API_KEY environment variable).",
    )
    _ = parser.add_argument(
        "-o",
        "--output",
        default="openai_utility_rates.json",
        help="Path to write the fetched rates (default: %(default)s).",
    )
    args = parser.parse_args(list(argv) if argv is not None else None)

    api_key = cast(str | None, args.api_key)
    if not api_key:
        parser.error("API key must be provided via --api-key or OPENEI_API_KEY environment variable.")

    rates_for_utility = cast(str | None, args.ratesforutility)

    if rates_for_utility is None:
        parser.error("ratesforutility must be set")

    effective_on_date = int(datetime.datetime.now(datetime.UTC).timestamp())
    records = list(
        iter_utility_rates(
            api_key=api_key,
            format="json",
            version="latest",
            ratesforutility=rates_for_utility,
            detail="minimal",
            sector="Residential",
            effective_on_date=effective_on_date,
        )
    )

    output_path = Path(cast(str, args.output))
    _ = output_path.write_text(json.dumps(records, indent=2))
    print(f"Wrote {len(records)} records to {output_path}")


if __name__ == "__main__":
    main()



class CoreEIA861_ASSN_UTILITY:
    https = "https://s3.us-west-2.amazonaws.com/pudl.catalyst.coop/nightly/core_eia861__assn_utility.parquet"
    s3 = "s3://pudl.catalyst.coop/nightly/core_eia861__assn_utility.parquet"
    s3_region = "us-west-2"

    class Columns:
        report_date = "report_date"
        "date, Date reported."
        state = "state"
        "string, Two letter US state abbreviation."
        utility_id_eia = "utility_id_eia"
        "integer, The EIA Utility Identification number."


class Core_PUDL_ASSN_EIA_PUDL_UTILITIES:
    """Association table providing connections between EIA utility IDs and manually assigned PUDL utility IDs."""

    https = "https://s3.us-west-2.amazonaws.com/pudl.catalyst.coop/nightly/core_pudl__assn_eia_pudl_utilities.parquet"
    s3 = "s3://pudl.catalyst.coop/nightly/core_pudl__assn_eia_pudl_utilities.parquet"

    class Columns:
        utility_id_eia = "utility_id_eia"
        "integer, primary key, The EIA Utility Identification number."
        utility_id_pudl = "utility_id_pudl"
        "integer, A manually assigned PUDL utility ID. May not be stable over time."
        utility_name_eia = "utility_name_eia"
        "string, The name of the utility."


class EIA_Yearly_Sales:
    """Annual time series of utilities electric sales from all rate schedules in effect throughout the year.

    https://catalystcoop-pudl.readthedocs.io/en/latest/data_dictionaries/pudl_db.html#core-eia861-yearly-sales
    """

    https = "https://s3.us-west-2.amazonaws.com/pudl.catalyst.coop/nightly/out_ferc1__yearly_sales_by_rate_schedules_sched304.parquet"
    s3 = "s3://pudl.catalyst.coop/nightly/out_ferc1__yearly_sales_by_rate_schedules_sched304.parquet"


class CORE_EIA861_Yearly_Sales:
    """Annual time series of electricity sales to ultimate customers by utility, balancing authority, state, and customer class.

    https://catalystcoop-pudl.readthedocs.io/en/latest/data_dictionaries/pudl_db.html#core-eia861-yearly-sales
    """

    https = "https://s3.us-west-2.amazonaws.com/pudl.catalyst.coop/nightly/core_eia861__yearly_sales.parquet"
    s3 = "s3://pudl.catalyst.coop/nightly/core_eia861__yearly_sales.parquet"



from datetime import date, datetime
from typing import Annotated, NotRequired

from pydantic import BeforeValidator, FailFast
from pydantic.alias_generators import to_camel
from typing_extensions import TypedDict

from .common import (
    CustomerClass,
    TariffChargePeriod,
    TariffChargeType,
    TariffEffectiveOnRule,
    TariffPrivacy,
    TariffServiceType,
    TariffType,
)
from .tariffproperty import TariffPropertyStandard
from .tariffrate import TariffRateExtended, TariffRateStandard


class TariffStandardFields(TypedDict):
    prior_tariff_id: NotRequired[int]
    "Unique Arcadia ID that identifies the prior revision of the tariffId above"
    distribution_lse_id: NotRequired[int | None]
    """
    In states like Texas where the load-serving entity that sells the power is
    different than the load-serving entity that distributes the power, this will
    contain the ID of the distribution LSE. Otherwise, it will be None.
    """
    tariff_type: TariffType
    """
    Possible values are:
      DEFAULT - tariff that is automatically given to this service class
      ALTERNATIVE - opt-in alternate tariff for this service class
      OPTIONAL_EXTRA - opt-in extra, such as green power or a smart thermostat program
      RIDER - charge that can apply to multiple tariffs. Often a regulatory-mandated charge
    """
    customer_class: CustomerClass | None
    """
    Possible values are:
      RESIDENTIAL - homes, apartments etc.
      GENERAL - commercial, industrial, and other business and organization service types
        (often have additional applicability criteria)
      SPECIAL_USE - examples are government, agriculture, street lighting, transportation
      PROPOSED - Utility rates that have been proposed by utilities and approved by utility commissions,
        but are not yet effective (requires product subscription)
    """
    customer_count: NotRequired[int]
    "Number of customers that are on this master tariff"
    """
    The likelihood that a customer is on this tariff of all the tariffs in the search results.
    """
    customer_likelihood: NotRequired[float | None]
    """
    The likelihood that a customer is on this tariff of all the tariffs in the search results.
    Only populated when getting more than one tariff.
    """
    territory_id: int
    """
    ID of the territory that this tariff applies to.
    This is typically the service area for the LSE in this regulatory region (i.e. a state in the USA)
    """
    effective_date: date
    "Date on which this tariff is no longer effective. Can be null which means end date is not known or tariff is open-ended"
    end_date: date | None
    "Date on which the tariff was or will be effective"
    time_zone: str
    "If populated (usually is), it's the timezone that this tariffs dates and times refer to"
    effective_on_rule: NotRequired[TariffEffectiveOnRule]
    billing_period: str
    "How frequently bills are generated."
    currency: str
    "ISO Currency code that the rates for this tariff refer to (e.g. USD for USA Dollar)"
    charge_types: Annotated[
        list[TariffChargeType], BeforeValidator(lambda _: _.split(",") if isinstance(_, str) else _)  # pyright: ignore[reportAny]
    ]
    "List of all the different ChargeType rates on this tariff"
    charge_period: TariffChargePeriod
    "The most fine-grained period for which charges are calculated."
    has_time_of_use_rates: bool
    "Indicates whether this tariff contains one or more Time of Use Rate."
    has_tiered_rates: bool
    "Indicates whether this tariff contains one or more Tiered Rate."
    has_contracted_rates: bool
    """
    Indicates whether this tariff contains one or more Rate that can be contracted
    (sometimes called by-passable or associated with a price to compare).
    """
    has_rate_applicability: bool
    """
    Indicates that one or more rates on this tariff are only applicable to customers with a particular circumstance.
    When true, this will be specified in the TariffProperty collection, and also on the TariffRate or rates in question.
    """


class TariffExtendedFields(TypedDict):
    tariff_book_name: str
    "Name of the tariff as it appears in the tariff document"
    lse_code: str
    "Abbreviated name of the load-serving entity"
    customer_count_source: NotRequired[str]
    "Where we got the customer count numbers from. Typically FERC (form 1 filings) or Arcadia (our own estimates)."
    closed_date: datetime | None
    """
    Date on which a tariff became closed to new customers,
    but still available for customers who were on it at the time.
    Can be null which means that the tariff is not closed.
    All versions of a particular tariff
    (i.e. those that share a particular masterTariffId)
    will have the same closedDate value.
    """
    min_monthly_consumption: float | None
    "When applicable, the minimum monthly consumption allowed to be eligible for this tariff."
    "When applicable, the maximum monthly consumption allowed to be eligible for this tariff."
    max_monthly_consumption: float | None
    "When applicable, the minimum monthly demand allowed to be eligible for this tariff."
    min_monthly_demand: float | None
    "When applicable, the maximum monthly demand allowed to be eligible for this tariff."
    max_monthly_demand: float | None
    "Indicates that this tariff has additional eligibility criteria, as specified in the TariffProperty collection"
    has_tariff_applicability: bool
    has_net_metering: bool
    "Indicates whether this tariff contains one or more net metered rates."
    privacy: TariffPrivacy
    "Privacy status of the tariff."


class TariffMinimalFields(TypedDict):
    is_active: bool

    tariff_id: int
    "Unique Arcadia ID (primary key) for this tariff"
    master_tariff_id: int
    "Unique Arcadia ID that persists across all revisions of this tariff"
    tariff_code: str
    "Shortcode that the LSE uses as an alternate name for the tariff"
    tariff_name: str
    "Name of the tariff as used by the LSE"
    lse_id: int
    "ID of load serving entity this tariff belongs to"
    lse_name: str
    "Name of the load-serving entity"
    service_type: NotRequired[TariffServiceType]
    "Type of service for the tariff"


class TariffMinimal(TariffMinimalFields):
    __pydantic_config__ = {"alias_generator": to_camel, "extra": "forbid"}  # pyright: ignore[reportGeneralTypeIssues, reportUnannotatedClassAttribute]  # noqa


class TariffStandard(TariffMinimalFields, TariffStandardFields):
    __pydantic_config__ = {"alias_generator": to_camel, "extra": "forbid"}  # pyright: ignore[reportGeneralTypeIssues, reportUnannotatedClassAttribute]  # noqa
    properties: NotRequired[list[TariffPropertyStandard]]
    rates: NotRequired[list[TariffRateStandard]]


class TariffExtended(TariffMinimalFields, TariffStandardFields, TariffExtendedFields):
    __pydantic_config__ = {"alias_generator": to_camel, "extra": "forbid"}  # pyright: ignore[reportGeneralTypeIssues, reportUnannotatedClassAttribute]  # noqa
    properties: NotRequired[list[TariffPropertyStandard]]
    rates: NotRequired[Annotated[list[TariffRateExtended], FailFast()]]


Tariff = TariffMinimal | TariffStandard | TariffExtended



from datetime import date
import os
from pprint import pprint

from dotenv import load_dotenv

from tariff_fetch.rateacuity import LoginState, create_context
from tariff_fetch.urdb.rateacuity_history_gas import build_urdb


def main():
    _ = load_dotenv()
    username = os.getenv("RATEACUITY_USERNAME")
    password = os.getenv("RATEACUITY_PASSWORD")
    with create_context() as context:
        state = (
            LoginState(context)
            .login(username or "", password or "")
            .gas()
            .history()
            .select_state("NY")
            .select_utility("Corning Natural Gas")
            .select_schedule("SC 14-AGGREGATE GROUP FIRM TRANSPORTATION-Residential--")
            .set_enddate(date(2025, 12, 1))
            .set_number_of_comparisons(12)
            .set_frequency(1)
        )
        report = state.as_dataframe()
        state.back_to_selections()
        print(report)
        print(report.schema)
        pprint(build_urdb(report))


def main_():
    _ = load_dotenv()
    username = os.getenv("RATEACUITY_USERNAME")
    password = os.getenv("RATEACUITY_PASSWORD")
    with create_context() as context:
        report = (
            LoginState(context)
            .login(username or "", password or "")
            .gas()
            .history()
            .select_state("RI")
            .select_utility("Rhode Island Energy (formally National Grid)")
            .select_schedule("12-RESIDENTIAL HEATING---")
            .set_enddate(date(2025, 12, 1))
            .set_number_of_comparisons(12)
            .set_frequency(1)
            .as_dataframe()
        )
        print(report)
        print(report.schema)
        print(build_urdb(report))


if __name__ == "__main__":
    main()



import pytest

from tariff_fetch.urdb.rateacuity_history_gas import (
    _build_energy_schedule_raw,
    _build_static_charges,
    _get_monthly_taxes,
    _seasonal_month_fraction,
)
from tariff_fetch.urdb.rateacuity_history_gas.history_data import (
    ConsumptionRow,
    DayOfMonth,
    FixedChargeRow,
    PercentageRow,
    Season,
)
from tariff_fetch.urdb.rateacuity_history_gas.shared import kwh_multiplier


def test_seasonal_month_fraction_handles_partial_wraparound_seasons() -> None:
    season = Season(
        start=DayOfMonth(day=15, month=11),
        end=DayOfMonth(day=10, month=4),
    )

    assert _seasonal_month_fraction(season, 2025, 9) == 0
    assert _seasonal_month_fraction(season, 2025, 10) == pytest.approx(16 / 30)
    assert _seasonal_month_fraction(season, 2025, 11) == 1
    assert _seasonal_month_fraction(season, 2025, 3) == pytest.approx(10 / 30)
    assert _seasonal_month_fraction(season, 2025, 4) == 0


def test_seasonal_month_fraction_uses_actual_year_for_leap_february() -> None:
    season = Season(
        start=DayOfMonth(day=1, month=2),
        end=DayOfMonth(day=29, month=2),
    )

    assert _seasonal_month_fraction(season, 2024, 1) == 1
    assert _seasonal_month_fraction(season, 2025, 1) == pytest.approx(28 / 28)


def test_build_helpers_prorate_percentage_and_fixed_rows_by_season() -> None:
    season = Season(
        start=DayOfMonth(day=15, month=11),
        end=DayOfMonth(day=10, month=12),
    )
    percentage_row = PercentageRow(
        rate="Tax",
        season=season,
        year=2025,
        effective_date=None,
        month_values=[10.0] * 12,
        location_avg_factor=1,
        rate_determinant="percent",
    )
    fixed_row = FixedChargeRow(
        rate="Customer Charge",
        season=season,
        year=2025,
        effective_date=None,
        month_values=[30.0] * 12,
        location_avg_factor=1,
        rate_determinant="per month",
    )

    monthly_taxes = _get_monthly_taxes([percentage_row])
    static_charges = _build_static_charges([fixed_row])
    fixed_charge = static_charges.get("fixedchargefirstmeter")

    assert monthly_taxes[9] == 1
    assert monthly_taxes[10] == pytest.approx(1 + 0.1 * (16 / 30))
    assert monthly_taxes[11] == pytest.approx(1 + 0.1 * (10 / 31))
    assert fixed_charge is not None
    assert fixed_charge == pytest.approx((30 * (16 / 30) + 30 * (10 / 31)) / 12)


def test_build_energy_schedule_prorates_consumption_rows_by_season() -> None:
    season = Season(
        start=DayOfMonth(day=15, month=11),
        end=DayOfMonth(day=10, month=12),
    )
    baseline_row = ConsumptionRow(
        rate="Base",
        season=None,
        year=2025,
        effective_date=None,
        month_values=[1.0] * 12,
        location_avg_factor=1,
        rate_determinant="per therm",
    )
    seasonal_row = ConsumptionRow(
        rate="Winter Adder",
        season=season,
        year=2025,
        effective_date=None,
        month_values=[1.0] * 12,
        location_avg_factor=1,
        rate_determinant="per therm",
    )

    urdb = _build_energy_schedule_raw([baseline_row, seasonal_row], include_taxes=False)
    rate_structure = urdb.get("energyratestructure")
    schedules = urdb.get("energyweekdayschedule")
    assert rate_structure is not None
    assert schedules is not None

    october_rate = rate_structure[schedules[9][0]][0]["rate"]
    november_rate = rate_structure[schedules[10][0]][0]["rate"]
    december_rate = rate_structure[schedules[11][0]][0]["rate"]

    base_rate = round(1 / kwh_multiplier("per therm"), 6)
    assert october_rate == pytest.approx(base_rate)
    assert november_rate == pytest.approx(round((1 / kwh_multiplier("per therm")) * (1 + 16 / 30), 6))
    assert december_rate == pytest.approx(round((1 / kwh_multiplier("per therm")) * (1 + 10 / 31), 6))
